from .base import BaseFinanceOverviewTransformer

class AgentOverviewTransformer(BaseFinanceOverviewTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
    
    def process_transform(self, tech_data, balance_sheet_list, profit_lose_list, cash_flow_list):
        """
        tech_data: tech_fetcher回傳的技術面資料 (pd.DataFrame)
        seasonal_data: 一系列的balance_sheet, profit_lose, cash_flow
        """
        seasonal_data = [
            {
                **balance_sheet['balance_sheet'],
                **profit_lose['income_statement'],
                **cash_flow['cash_flow']
            } for balance_sheet, profit_lose, cash_flow in zip(
                balance_sheet_list,
                profit_lose_list,
                cash_flow_list
            )
        ]

        TTM_data = self._get_TTM_data(seasonal_data)

        return_dict = {
            "date":
            self._get_latest_date(tech_data),
            "ticker":
            self.ticker,
            "company_name":self.company_name,
            "category":
            self._get_category(),
            "market_open":
            self._is_us_market_open(),
            "latest_close":
            self._get_latest_close(tech_data),
            "close_offset":
            self._get_latest_offset(tech_data).get("value", "0.0"),
            "close_offset_percentage":
            self._get_latest_offset(tech_data).get("percentage", "0.0%"),
            "latest_volume":
            self._get_latest_volume(tech_data),
            "average_volume":
            self._get_latest_average_volume(tech_data, 30),
            "market_capitalzation":
            self._get_market_capitalization(tech_data, seasonal_data),
            "P_E":
            self._get_PE(tech_data, TTM_data),
            "P_S":
            self._get_PS(tech_data, TTM_data, seasonal_data),
        }

        return return_dict
        


import abc
from datetime import datetime, timedelta, date
import json
import pandas as pd
from pymongo import MongoClient
import pytz
from typing import Union
from ..utils import StatsDateTime, StatsProcessor, YoY_Calculator


class StatsFetcher(abc.ABC):

    def __init__(self, ticker: str, db_client: MongoClient):
        self.ticker = ticker
        self.timezone = pytz.timezone("Asia/Taipei")
        self.tw_company_list = StatsProcessor.load_json("company_list/tw.json")
        db_mapping = {
            "company": "twse_stats",
            "company_us": "us_stats",
        }

        name_mapping = {"company": "台股", "company_us": "美股"}

        db_name = "company" if self.ticker in self.tw_company_list else "company_us"
        self.db = db_client[db_name]
        self.collection_name = db_mapping.get(db_name, "unknown")
        assert self.collection_name != "unknown", f"請確認 {ticker} 是否是 {','.join(list(name_mapping.values()))}"
        self.collection = db_client[db_name][self.collection_name]

        self.target_metric_dict = {
            'value': ['value'],
            'value_and_percentage': ['value', 'percentage'],
            'percentage': ['percentage'],
            'grand_total': ['grand_total'],
            'grand_total_values': ['grand_total', 'grand_total_percentage'],
            'grand_total_percentage': ['grand_total_percentage'],
            'growth': [f'YoY_{i}' for i in [1, 3, 5, 10]],
            'grand_total_growth': [f"YoY_{i}" for i in [1, 3, 5, 10]]
        }

    def prepare_query(self):
        return [
            {
                "$match": {
                    "ticker": self.ticker,
                }
            },
        ]
    
    def query_data(self):
        return NotImplementedError()

    def collect_data(self):
        pipeline = self.prepare_query()
        fetched_data = list(self.collection.aggregate(pipeline))
        return fetched_data if fetched_data else None

    def str_to_datetime(self, date_str: str) -> StatsDateTime:
        date = self.timezone.localize(datetime.strptime(date_str, "%Y-%m-%d"))
        year, month, day = date.year, date.month, date.day
        season = (month - 1) // 3 + 1
        return StatsDateTime(date, year, month, day, season)

    def has_required_columns(
        self, df: pd.DataFrame, required_cols=None
    ) -> bool:
        if required_cols is None:
            required_cols = ['date', 'open', 'high', 'low', 'close', 'volume']
        return all(col in df.columns for col in required_cols)
    
    @staticmethod
    def flatten_dict(value_dict, indexes, target_keys):
        indexes = value_dict.keys()
        new_dict = {}

        for key in indexes:
            new_dict.update(
                {
                    f"{key}_{sub_key}": value_dict[key].get(sub_key, None) 
                    for sub_key in target_keys
                }
            )
        return new_dict


class BaseTEJFetcher(abc.ABC):

    @abc.abstractmethod
    def get(self):
        pass

    def get_latest_data_time(self, ticker):
        latest_data = self.collection.find_one(
            {"ticker": ticker}, {
                "last_update": 1,
                "_id": 0
            }
        )

        if (latest_data):
            # return 得到最新日期或None
            return latest_data.get('last_update', {}).get("latest_data_date", None)
        
        else:
            return None

    def process_value(self, value):
        if isinstance(value, str) and "%" in value:
            value = value.replace("%", "")
        try:
            return float(value)
        except (ValueError, TypeError):
            return None

    def calculate_growth(self, this_value, last_value, delta):
        try:
            return YoY_Calculator.cal_growth(
                this_value, last_value, delta
            ) * 100
        except Exception:
            return None

    def cal_YoY(
        self, data_dict: dict, start_year: int, end_year: int, season: int
    ):
        year_shifts = [1, 3, 5, 10]
        return_dict = {}

        for year in range(start_year, end_year + 1):
            year_data = data_dict.get(f"{year}Q{season}", {}).copy()
            if not year_data:
                continue

            for key, value in list(year_data.items()):
                if key == "season":
                    continue

                this_value = self.process_value(value)
                if this_value is None:
                    year_data.pop(key)
                    continue

                temp_dict = {"value": value}
                for shift in year_shifts:
                    past_value = self.process_value(
                        data_dict.get(f"{year - shift}Q{season}", {}).get(key)
                    )
                    growth = self.calculate_growth(this_value, past_value, shift) if past_value else None

                    temp_dict[
                        f"YoY_{shift}"
                    ] = f"{growth:.2f}%" if growth else None

                year_data[key] = temp_dict
            return_dict[f"{year}Q{season}"] = year_data
        return return_dict

    def cal_QoQ(self, data_dict):
        return_dict = {}

        for time_index, this_data in data_dict.items():
            year, season = map(int, time_index.split("Q"))
            last_year, last_season = (
                year - 1, 4
            ) if season == 1 else (year, season - 1)

            for key in list(this_data.keys()):
                if key == "season":
                    continue

                this_value = self.process_value(this_data[key])
                if this_value is None:
                    this_data.pop(key)
                    continue

                temp_dict = {"value": this_data[key]}
                last_value = data_dict.get(
                    f"{last_year}Q{last_season}",{}
                ).get(key, {}).get('value')

                last_value = self.process_value(last_value)
                growth = self.calculate_growth(
                    this_value, last_value, 1
                ) if last_value is not None else None
                temp_dict['growth'] = (f"{growth:.2f}%" if growth else None)

                this_data[key] = temp_dict

            return_dict[time_index] = this_data

        return return_dict

    def get_dict_of_df(self, data_dict):
        """
        dict[dict] -> dict[df]
        """
        return {
            key: pd.DataFrame.from_dict(data) for key, data in data_dict.items()
        }

    def set_time_shift(self, date: Union[str, datetime], period: str):
        if isinstance(date, str):
            date = datetime.strptime(date, "%Y-%m-%d")

        period_mapping = {
            "1d": timedelta(days=1),
            "7d": timedelta(days=7),
            "1m": timedelta(days=30),
            "3m": timedelta(days=90),
            "1y": timedelta(days=365),
            "3y": timedelta(days=365 * 3),
            "5y": timedelta(days=365 * 5),
            "10y": timedelta(days=365 * 10),
        }

        if period == "all":
            return datetime.strptime("1991-01-01", "%Y-%m-%d")

        return date - period_mapping.get(period, timedelta(days=0))    # 預設為不變"



from .base import AsyncBaseFetcher
from datetime import datetime
from ..db_extractors import (
    AsyncYFDailyTechDBExtractor, AsyncTEJDailyTechDBExtractor,
    AsyncBatchTechDBExtractor
)
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import DailyTechTransformer, BatchTechTransformer
from neurostats_API.utils import NotSupportedError
import pandas as pd
from pytz import timezone
from neurostats_API.utils import StatsDateTime
import yfinance as yf
from pymongo import AsyncMongoClient

class AsyncTechFetcher(AsyncBaseFetcher):

    def __init__(self, ticker, client):
        self.ticker = ticker
        self.transformer_map = {
            "tw": DailyTechTransformer,
            "us": DailyTechTransformer
        }
        try:
            self.yf_db_extractor = get_extractor("YF_tech", ticker, client)

            self.tej_db_extractor = None
            if (self.yf_db_extractor.get_zone() == 'tw'):
                self.tej_db_extractor = get_extractor("TEJ_tech", ticker, client)

            self.company_name = self.yf_db_extractor.get_company_name()
            self.zone = self.yf_db_extractor.get_zone()

            self.required_cols = [
                'date', 'open', 'high', 'low', 'close', 'volume'
            ]

            transformer = self.get_transformer(self.zone)
            self.transformer = transformer(ticker, self.company_name, self.zone)

        except NotSupportedError as e:
            raise NotSupportedError(
                f"{self.__class__.__name__} only supports {list(self.transformer_map.keys())}, got {self.zone} with ticker = \"{self.ticker}\""
            ) from e

    async def query_data(self, start_date=None, end_date=None):
        if (start_date is None):
            start_date = datetime.strptime("1991-01-01", "%Y-%m-%d")
        if (end_date is None):
            end_date = datetime.today()

        fetched_data = await self._fetch_data(start_date, end_date)

        transformed_data = self.transformer.process_transform(fetched_data)

        return transformed_data

    async def get_daily(self, start_date=None, end_date=None):
        await self.query_data(start_date, end_date)

        return self.transformer.get_daily()

    async def get_weekly(self, start_date=None, end_date=None):
        await self.query_data(start_date, end_date)

        return self.transformer.get_weekly()

    async def get_monthly(self, start_date=None, end_date=None):
        await self.query_data(start_date, end_date)

        return self.transformer.get_monthly()

    async def get_quarterly(self, start_date=None, end_date=None):
        await self.query_data(start_date, end_date)

        return self.transformer.get_quarterly()

    async def get_yearly(self, start_date=None, end_date=None):
        await self.query_data(start_date, end_date)

        return self.transformer.get_yearly()

    async def _fetch_data(self, start_date, end_date):

        if (self.ticker in ['^GSPC', '^IXIC', '^DJI', '^TWII']):
            df = self._conduct_yf_search(self.ticker, start_date, end_date)

            return df[self.required_cols]

        else:
            search_fn_map = {
                'tw': self._conduct_tw_db_search,
                'us': self._conduct_us_db_search
            }
            search_fn = search_fn_map.get(self.zone)
            df = await search_fn(start_date, end_date)

            return df

    def _conduct_yf_search(self, ticker, start_date, end_date):
        yf_ticker = yf.Ticker(ticker)
        origin_df = yf_ticker.history(period="max")

        if origin_df.empty:
            return origin_df

        start_date = StatsDateTime.transform_date(start_date, tz=str(origin_df.index.tz)) # 跟隨yf搜尋結果的時區
        end_date = StatsDateTime.transform_date(end_date, tz=str(origin_df.index.tz))

        origin_df = origin_df.reset_index()
        origin_df["Date"] = pd.to_datetime(origin_df["Date"])
        df = origin_df.rename(
            columns={
                "Date": "date",
                "Open": "open",
                "High": "high",
                "Low": "low",
                "Close": "close",
                "Volume": "volume"
            }
        )

        df = df.loc[(df['date'] >= start_date) & (df['date'] <= end_date)]
        df['date'] = df['date'].dt.tz_convert("Asia/Taipei")
        return df

    async def _conduct_tw_db_search(self, start_date, end_date):
        search_fns = [
            self._conduct_tw_db_search_yf,
            self._conduct_tw_db_search_tej,
        ]

        sync_search_fns = [
            lambda s_date, e_date: self.
            _conduct_yf_search(f'{self.ticker}.tw', s_date, e_date),
            lambda s_date, e_date: self.
            _conduct_yf_search(f'{self.ticker}.two', s_date, e_date),
        ]

        for search_method in search_fns:
            try:
                fetched_data = await search_method(start_date, end_date)
                if not fetched_data.empty:
                    break
            except (KeyError, ValueError, TypeError) as e:
                print(e)
                continue
        else:    # 找不到資料
            for search_method in sync_search_fns:
                fetched_data = search_method(start_date, end_date)
                if not fetched_data.empty:
                    break
            else:
                return pd.DataFrame(self.required_cols)

        return fetched_data[self.required_cols]

    async def _conduct_tw_db_search_yf(self, start_date, end_date):
        """
        尋找從yf與twse爬下來的台股資料
        """
        fetched_data = await self.yf_db_extractor.query_data(
            start_date, end_date
        )

        if (fetched_data):
            df = pd.DataFrame(fetched_data)
            return df
        else:
            return None

    async def _conduct_tw_db_search_tej(self, start_date, end_date):
        """
        尋找從TEJ爬下來的台股資料
        """
        fetched_data = await self.tej_db_extractor.query_data(
            start_date, end_date
        )

        if (fetched_data):
            df = pd.DataFrame(fetched_data)
            df = df.rename(
                columns={
                    "open_d": "open",
                    "high_d": "high",
                    "low_d": "low",
                    "close_d": "close",
                    "vol": "volume"
                }
            )

            return df
        else:
            return None

    def _conduct_us_db_search(self, start_date, end_date):
        search_fns = [
            self.yf_db_extractor.query_data, lambda s_date, e_date: self.
            _conduct_yf_search(f"{self.ticker}", s_date, e_date)
        ]

        for search_method in search_fns:
            try:
                df = search_method(start_date, end_date)
                break
            except (KeyError, ValueError, TypeError):
                continue
        else:    # 回傳空df
            df = pd.DataFrame(columns=self.required_cols)

        return df

class AsyncBatchTechFetcher(AsyncBaseFetcher):
    def __init__(
            self, client:AsyncMongoClient, tickers:list, 
            zone:str, strftime: str = "%Y-%m-%d"
    ):
        super().__init__()
        self.tickers = tickers
        self.zone = zone
        self.extractor = AsyncBatchTechDBExtractor(
            ticker=self.tickers, client=client, zone=zone
        )
        self.transformer = BatchTechTransformer(strftime=strftime)

    async def query_data(
        self, start_date=None, end_date=None, get_latest=False,
    ):
        fetched_data = await self.extractor.query_data(
            start_date, end_date, get_latest,
        )

        transformed_data = self.transformer.process_transform(
            fetched_datas = list(fetched_data),
        )

        return transformed_data


from .daily import (
    AsyncTEJDailyChipDBExtractor,
    AsyncTEJDailyTechDBExtractor,
    AsyncYFDailyTechDBExtractor,
    AsyncDailyValueDBExtractor,
    AsyncTEJDailyValueDBExtractor,
    AsyncTWSEChipDBExtractor,
    AsyncUS_F13DBExtractor,
    AsyncUS_CelebrityDBExtractor,
    AsyncBatchTechDBExtractor
)

from .month_revenue import (
    AsyncTWSEMonthlyRevenueExtractor
)

from .seasonal import (
    AsyncBalanceSheetExtractor,
    AsyncProfitLoseExtractor,
    AsyncCashFlowExtractor,
    AsyncTEJFinanceStatementDBExtractor,
    AsyncTEJSelfSettlementDBExtractor
)


from .base import AsyncBaseFetcher
from ..db_extractors import AsyncCashFlowExtractor
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import (
    TWSECashFlowTransformer, USCashFlowTransformer
)
from neurostats_API.utils import NoCompanyError, NotSupportedError


class AsyncCashFlowFetcher(AsyncBaseFetcher):

    def __init__(self, ticker, client):

        super().__init__()
        self.transformer_map = {
            "tw": TWSECashFlowTransformer,
            "us": USCashFlowTransformer
        }
        self.ticker = ticker
        try:
            self.extractor = get_extractor("DB_cash_flow", self.ticker, client)
            self.company_name = self.extractor.get_company_name()
            self.zone = self.extractor.get_zone()

            transformer = self.get_transformer(self.zone)
            self.transformer = transformer(ticker, self.company_name, self.zone)
        
        except NotSupportedError as e:
            raise NotSupportedError(
                f"{self.__class__.__name__} only support {list(self.transformer_map.keys())} companies now, {ticker} is not available"
            ) from e
        except NoCompanyError:
            raise

    async def query_data(self, start_date=None, end_date=None):
        fetched_data = await self.extractor.query_data(start_date, end_date)

        transformed_data = self.transformer.process_transform(fetched_data)

        return transformed_data

    async def query_data_stats_only(self, start_date=None, end_date=None):
        fetched_data = await self.extractor.query_data(start_date, end_date)

        transformed_data = self.transformer.process_stats_page(fetched_data)

        return transformed_data

    async def query_data_QoQ_and_YoY_only(self, start_date=None, end_date=None):
        if (self.zone != 'tw'):
            raise NotSupportedError("query_data_QoQ_and_YoY_only() only supports for TW companies")
        fetched_data = await self.extractor.query_data(start_date, end_date)

        transformed_data = self.transformer.process_QoQ(fetched_data)

        return transformed_data



import abc
from neurostats_API.utils import StatsProcessor, YoY_Calculator
import pandas as pd


class BaseTransformer(abc.ABC):
    """
    Transformer用途: 轉換資料為fetcher可使用的格式
    """

    def __init__(self, ticker, company_name, zone):
        self.ticker = ticker
        self.company_name = company_name

        self.zone = zone
        self.return_keys = []

    @abc.abstractmethod
    def process_transform(self):
        pass

    @classmethod
    def flatten_dict(cls, value_dict, target_keys):
        indexes = value_dict.keys()
        new_dict = {}

        for key in indexes:
            new_dict.update(
                {
                    f"{key}_{sub_key}": value_dict[key].get(sub_key, None)
                    for sub_key in target_keys
                }
            )
        
        return new_dict

    @staticmethod
    def _process_unit(data_df, postfix):

        lambda_map = {
            "_value":
            lambda x: StatsProcessor.cal_non_percentage(x, postfix="千元"),
            "_percentage":
            lambda x: StatsProcessor.
            cal_non_percentage(x, to_str=True, postfix="%"),
            '_growth':
            lambda x: StatsProcessor.
            cal_non_percentage(x, to_str=True, postfix="%"),
            "_YoY_1":
            StatsProcessor.cal_percentage,
            "_YoY_3":
            StatsProcessor.cal_percentage,
            "_YoY_5":
            StatsProcessor.cal_percentage,
            "_YoY_10":
            StatsProcessor.cal_percentage
        }

        process_fn = lambda_map.get(postfix)
        postfix_cols = data_df.columns.str.endswith(postfix)
        postfix_cols = data_df.loc[:, postfix_cols].columns

        if (postfix == "_value"):
            postfix_cols = [
                col for col in postfix_cols
                if not ("eps" in col or "每股盈餘" in col)
            ]

        if (postfix == '_growth'):
            data_df[postfix_cols] = data_df[postfix_cols].map(
                lambda x: x * 100.0 if isinstance(x, float) else x
            )

        data_df[postfix_cols] = (data_df[postfix_cols].map(process_fn))

        return data_df

    def _apply_process_unit_pipeline(
        self, data_df, postfix_list=["_value", "percentage"]
    ):
        for postfix in postfix_list:
            data_df = self._process_unit(data_df, postfix)
        return data_df

    @staticmethod
    def _slice_target_season(stats_df, target_season):
        target_season_columns = stats_df.columns.str.endswith(
            f"Q{target_season}"
        )
        stats_df = stats_df.loc[:, target_season_columns]

        return stats_df

    def _get_empty_structure(self):
        return_dict = {
            "warning": "No data fetched",
            "ticker": self.ticker,
            "company_name": self.company_name
        }

        return_dict.update(
            {
                key: pd.DataFrame(columns=pd.Index([], name='index'))
                for key in self.return_keys
            }
        )
        return return_dict

    def _process_value(self, value):
        if isinstance(value, str) and "%" in value:
            value = value.replace("%", "")
        try:
            return float(value)
        except (ValueError, TypeError):
            return None

    def _calculate_growth(self, this_value, last_value, delta):
        try:
            return YoY_Calculator.cal_growth(
                this_value, last_value, delta
            ) * 100
        except Exception:
            return None

    def _cal_QoQ(self, data_dict):
        """
        data_dict: {"<key>_value": {"2020Q1": <value>, ....}, "<key>_percentage": {"2020Q1": <value>, ....}}
        """

        return_dict = data_dict.copy()

        for key, datas in data_dict.items():
            if (key.endswith("_value")):
                main_key = key.split("_")[0]
                temp_growth_dict = dict()
                for year_season, value in datas.items():

                    year, season = year_season.split('Q')
                    year, season = int(year), int(season)

                    last_year, last_season = (
                        year, season - 1
                    ) if season != 1 else (year - 1, 4)

                    last_value = datas.get(f"{last_year}Q{last_season}", None)

                    growth = YoY_Calculator.cal_growth(
                        value, last_value, delta=1
                    )
                    if (growth):
                        growth = growth * 100
                        growth = f"{growth:.2f}%"

                    temp_growth_dict[year_season] = growth

                return_dict[f"{main_key}_growth"] = growth
        return_df = pd.DataFrame.from_dict(return_dict)
        return return_df



from .base import BaseDBClient
import os
from pymongo import AsyncMongoClient
from dotenv import load_dotenv

class TWSEDBClient(BaseDBClient):
    def __init__(self, client):
        super().__init__(client, zone= "tw")


from .datetime import StatsDateTime
from .db_client import DBClient
from .data_process import StatsProcessor
from .calculate_value import YoY_Calculator
from .exception import NoCompanyError, NoDataError, NotSupportedError



import logging
from functools import wraps

class StatsLogger:

    @classmethod
    def exception_log(cls, default_return=None, level='error'):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    log_fn = getattr(logging, level, logging.error)
                    log_fn(
                        f"{func.__name__} failed | args={args}, kwargs={kwargs} | error={e}",
                        exc_info=True
                    )
                    return default_return
            return wrapper
        return decorator


from .base import AsyncBaseFetcher
from ..db_extractors import AsyncTWSEMonthlyRevenueExtractor
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import TWSEMonthlyRevenueTransformer
from neurostats_API.utils import NoCompanyError, NotSupportedError


class AsyncMonthlyRevenueFetcher(AsyncBaseFetcher):

    def __init__(self, ticker, client):
        self.ticker = ticker
        self.transformer_map = {
            "tw": TWSEMonthlyRevenueTransformer,
        }

        try:
            self.extractor = get_extractor("TWSE_month_revenue", ticker, client)
            company_name = self.extractor.get_company_name()
            zone = self.extractor.get_zone()

            transformer = self.get_transformer(zone)

            self.transformer = transformer(ticker, company_name, zone)
        except NotSupportedError as e:
            raise NotSupportedError(
                "AsyncMonthRevenueFetcher only supports tw company now"
            ) from e


    async def query_data(self, start_date=None, end_date=None):
        fetched_data = await self.extractor.query_data(start_date, end_date)

        transformed_data = self.transformer.process_transform(fetched_data)

        return transformed_data



from .base import AsyncBaseSeasonalDBExtractor, StatsDateTime
from datetime import datetime
import json
import pandas as pd
from pymongo import ASCENDING, DESCENDING
from neurostats_API.async_mode.db import TWSEDBClient, USDBClient
from neurostats_API.utils import StatsDateTime, StatsProcessor
import yaml

class AsyncProfitLoseExtractor(AsyncBaseSeasonalDBExtractor):
    def __init__(self, ticker, client):
        super().__init__(ticker, client)

        self.column_name_map = {
            'tw': "profit_lose",
            'us': "income_statement"
        }


from .base import BaseDailyTechDBExtractor


class AsyncYFDailyTechDBExtractor(BaseDailyTechDBExtractor):

    def __init__(self, ticker, client):
        super().__init__(ticker, client)

    def _get_collection_name(self):
        self.collection_name_map = {"tw": "twse_daily_share_price", "us": "us_technical"}

        return self.collection_name_map.get(self.zone, None)



import pandas as pd


class InstitutionTradingProcessor:
    @classmethod
    def _process_latest_trading(cls, latest_trading):
        for key in latest_trading.keys():
            if (key.find("外陸資") >= 0 or key.find("外資") >= 0):
                cls._target_institution(
                    latest_trading, latest_table['foreign'], key, volume
                )
            elif (key.find("自營商") >= 0):
                cls._target_institution(
                    latest_trading, latest_table['prop'], key, volume
                )
            elif (key.find("投信") >= 0):
                cls._target_institution(
                    latest_trading, latest_table['mutual'], key, volume
                )
            elif (key.find("三大法人") >= 0):
                cls._target_institution(
                    latest_trading, latest_table['institutional_investor'], key,
                    volume
                )
        # 計算合計
        for unit in ['stock', 'percentage']:
            # 買進總和
            latest_table['institutional_investor']['buy'][unit] = (
                latest_table['foreign']['buy'][unit] +
                latest_table['prop']['buy'][unit] +
                latest_table['mutual']['buy'][unit]
            )
            # 賣出總和
            latest_table['institutional_investor']['sell'][unit] = (
                latest_table['foreign']['sell'][unit] +
                latest_table['prop']['sell'][unit] +
                latest_table['mutual']['sell'][unit]
            )

        frames = []
        for category, trades in latest_table.items():
            temp_df = pd.DataFrame(trades).T
            temp_df['category'] = category
            frames.append(temp_df)

        latest_df = pd.concat(frames)
        latest_df = latest_df.reset_index().rename(columns={'index': 'type'})
        latest_df = latest_df[[
            'type', 'category', 'stock', 'price', 'average_price', 'percentage'
        ]]

        latest_df = pd.melt(
            latest_df,
            id_vars=['type', 'category'],
            var_name='variable',
            value_name='value'
        )

        latest_df = latest_df.pivot_table(
            index=['category', 'variable'],
            columns='type',
            values='value',
            aggfunc='first'
        )

        # 重設列名，去除多層索引
        latest_df.columns.name = None    # 去除列名稱
        latest_df = latest_df.reset_index()

        return latest_df
    @classmethod
    def _process_annual_trading(cls, trading_datas):
        return pd.DataFrame(trading_datas)

    @classmethod
    def _target_institution(cls, old_table, new_table, key, volume):
        if (key.find("買進") >= 0):
            cls._cal_institution(old_table, new_table['buy'], key, volume)
        elif (key.find("賣出") >= 0):
            cls._cal_institution(old_table, new_table['sell'], key, volume)
        elif (key.find("買賣超") >= 0):
            cls._cal_institution(
                old_table, new_table['over_buy_sell'], key, volume
            )

    @classmethod
    def _cal_institution(cls, old_table, new_table, key, volume):
        new_table['stock'] = np.round(old_table[key] / 1000, 2).item()
        new_table['percentage'] = np.round((old_table[key] / volume) * 100,
                                           2).item()



from .base import StatsFetcher, StatsDateTime
from datetime import datetime
import json
import numpy as np
import pandas as pd
from pymongo import ASCENDING, DESCENDING
import pytz
import holidays
import warnings
from ..utils import StatsDateTime, StatsProcessor, NoCompanyError
import importlib.resources as pkg_resources
import yaml


class FinanceOverviewFetcher(StatsFetcher):
    """
    對應iFa.ai -> 財務分析 -> 重要指標(finance_overview)
    """

    def __init__(self, ticker, db_client):
        super().__init__(ticker, db_client)

        self.target_fields = StatsProcessor.load_yaml(
            "twse/finance_overview_dict.yaml"
        )
        self.inverse_dict = StatsProcessor.load_txt(
            "twse/seasonal_data_field_dict.txt", json_load=True
        )

    def prepare_query(self, target_year, target_season):

        pipeline = super().prepare_query()

        target_query = {
            "year": "$$target_season_data.year",
            "season": "$$target_season_data.season",
        }

        for key, target_sets in self.target_fields.items():
            try:
                small_targets = target_sets['field']

                value_index = target_sets['value']    # "金額" or "%"

                for small_target in small_targets:
                    big_target = self.inverse_dict[
                        small_target]    # balance_sheet/profit_lose/cash_flow
                    if (small_target == "利息費用_bank"):
                        small_target = small_target[:small_target.find("_bank")]
                    target_query.update(
                        {
                            f"{key}":
                            f"$$target_season_data.{big_target}.{small_target}.{value_index}"
                        }
                    )
            except Exception:
                continue

        pipeline.append(
            {
                "$project": {
                    "_id": 0,
                    "ticker": 1,
                    "company_name": 1,
                    "seasonal_data": {
                        "$map": {
                            "input": {
                                "$filter": {
                                    "input": "$seasonal_data",
                                    "as": "season",
                                    "cond": {
                                        "$and": [
                                            {
                                                "$eq":
                                                ["$$season.year", target_year]
                                            }, {
                                                "$eq": [
                                                    "$$season.season",
                                                    target_season
                                                ]
                                            }
                                        ]
                                    }
                                }
                            },
                            "as": "target_season_data",
                            "in": target_query
                        }
                    }
                }
            }
        )

        return pipeline

    def collect_data(self, target_year, target_season):
        pipeline = self.prepare_query(target_year, target_season)

        fetched_data = self.collection.aggregate(pipeline)

        fetched_data = list(fetched_data)

        return fetched_data[0]

    def query_data(self):

        try:
            latest_time = StatsDateTime.get_latest_time(
                self.ticker, self.collection
            )['last_update_time']
            year = latest_time['seasonal_data']['latest_year']
            season = latest_time['seasonal_data']['latest_season']
        except Exception as e:
            today = StatsDateTime.get_today()
            year = today.year - 1 if (today.season == 1) else today.year
            season = 4 if (today.season == 1) else today.season - 1

        fetched_data = self.collect_data(year, season)
        finance_dict = fetched_data['seasonal_data'][0]
        FinanceOverviewProcessor.process_rate(finance_dict)
        FinanceOverviewProcessor.process_all(finance_dict)
        self.fill_nan_index(finance_dict)
        FinanceOverviewProcessor.process_thousand_dollar(finance_dict)
        fetched_data['seasonal_data'] = finance_dict

        return fetched_data

    def fill_nan_index(self, finance_dict):
        for key in self.target_fields.keys():
            if (key not in finance_dict.keys()):
                finance_dict[key] = None


class FinanceOverviewProcessor(StatsProcessor):

    @classmethod
    def process_rate(cls, finance_dict):
        for key in finance_dict:
            if ('YoY' in key):
                finance_dict[key] = StatsProcessor.cal_percentage(
                    finance_dict[key]
                )
            elif ("rate" in key or 'ratio' in key):
                finance_dict[key] = StatsProcessor.cal_non_percentage(
                    finance_dict[key], to_str=True, postfix='%'
                )
            else:
                finance_dict[key] = StatsProcessor.cal_non_percentage(
                    finance_dict[key]
                )

    @classmethod
    def process_thousand_dollar(cls, finance_dict):
        process_index = [
            "revenue", "gross_profit", "operating_income", "net_income",
            "operating_cash_flow", "invest_cash_flow", "financing_cash_flow",
            "fcf", 'current_assets', 'current_liabilities',
            'non_current_assets', 'non_current_liabilities', 'total_assets',
            "total_liabilities", "equity"
        ]

        for index in process_index:
            try:
                finance_dict[index] = StatsProcessor.cal_non_percentage(
                    finance_dict[index], postfix="千元"
                )
            except Exception as e:
                finance_dict[index] = None

    @classmethod
    def process_all(cls, finance_dict):
        methods = [
            cls.cal_EBIT,
            cls.cal_share_outstanding,
            cls.cal_fcf,
            cls.cal_interest_bearing_debt,
            cls.cal_revenue_per_share,
            cls.cal_gross_per_share,
            cls.cal_operating_income_per_share,
            cls.cal_operating_cash_flow_per_share,
            cls.fcf_per_share,
            cls.cal_roa,
            cls.cal_roe,
            cls.cal_gross_over_asset,
            cls.cal_roce,
            cls.cal_gross_profit_marginal,
            cls.cal_operation_profit_rate,
            cls.cal_operating_cash_flow_profit_rate,
            cls.cal_dso,
            cls.cal_account_receive_over_revenue,
            cls.cal_dpo,
            cls.cal_inventories_cycle_ratio,
            cls.cal_dio,
            cls.cal_inventories_revenue_ratio,
            cls.cal_cash_of_conversion_cycle,
            cls.cal_asset_turnover,
            cls.cal_application_turnover,
            cls.cal_current_ratio,
            cls.cal_quick_ratio,
            cls.cal_debt_to_equity_ratio,
            cls.cal_net_debt_to_equity_ratio,
            cls.cal_interest_coverage_ratio,
            cls.cal_debt_to_operating_cash_flow,
            cls.cal_debt_to_free_cash_flow,
            cls.cal_cash_flow_ratio,
        ]

        for method in methods:
            method(finance_dict)

    @classmethod
    def cal_EBIT(cls, finance_dict):
        """
        計算EBIT
        EBIT = 營業收入 - 營業成本 - 營業費用 - 所得稅費用
        """
        try:
            EBIT = (
                finance_dict['revenue'] - finance_dict['operating_cost'] -
                finance_dict['operating_expenses'] - finance_dict['tax_fee']
            )
            finance_dict['EBIT'] = StatsProcessor.cal_non_percentage(EBIT)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['EBIT'] = None
            # print(f"Error calculating EBIT: {e}")

    @classmethod
    def cal_fcf(cls, finance_dict):
        """
        計算自由現金流(FCF):
        自由現金流 =  營業現金流 + 投資現金流
        """
        try:
            fcf = (
                finance_dict["operating_cash_flow"] +
                finance_dict["financing_cash_flow"]
            )
            finance_dict["fcf"] = StatsProcessor.cal_non_percentage(fcf)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['fcf'] = None
            # print(f"Error calculating FCF: {e}")

    @classmethod
    def cal_interest_bearing_debt(cls, finance_dict):
        """
        計算有息負債
        短期借款+長期借款
        """
        finance_dict['interest_bearing_debt'] = 0.0

        try:
            finance_dict['interest_bearing_debt'] += finance_dict[
                'short_term_liabilities']
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['interest_bearing_debt'] += 0.0
        try:
            finance_dict['interest_bearing_debt'] += finance_dict[
                'long_term_liabilities']
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['interest_bearing_debt'] += 0.0

    @classmethod
    def cal_share_outstanding(cls, finance_dict):
        """
        計算流通股數
        流通股數 = 本期淨利 ÷ 基本每股盈餘
        """
        try:
            finance_dict["share_outstanding"] = (
                finance_dict['net_income'] / finance_dict['eps']
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['share_outstanding'] = None
            # print(f"share_outstanding failed because of {str(e)}")

    @classmethod
    def cal_revenue_per_share(cls, finance_dict):
        """
        計算每股營收
        每股營收 = 營業收入 / 在外流通股數
        """
        try:
            revenue_per_share = (
                finance_dict['revenue'] / finance_dict['share_outstanding']
            )
            finance_dict['revenue_per_share'
                         ] = StatsProcessor.cal_non_percentage(
                             revenue_per_share, False
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['revenue_per_share'] = None
            # print(f"revenue_per_share failed because of {str(e)}")

    @classmethod
    def cal_gross_per_share(cls, finance_dict):
        """
        計算每股毛利
        = （當期營業毛利）÷（當期在外流通股數）
        """
        if ('gross_profit' not in finance_dict.keys()):
            try:
                finance_dict['gross_profit'] = (
                    finance_dict['revenue'] - finance_dict['operating_cost']
                )
            except:
                finance_dict['gross_profit'] = None
        try:
            gross_per_share = (
                finance_dict['gross_profit'] / finance_dict['share_outstanding']
            )
            finance_dict['gross_per_share'] = StatsProcessor.cal_non_percentage(
                gross_per_share, False
            )

        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['gross_per_share'] = None
            # print(f"gross_per_share failed because of {str(e)}")

    @classmethod
    def cal_operating_income_per_share(cls, finance_dict):
        """
        計算每股營業利益
        每股營業利益= （當期營業利益）÷（當期在外流通股數）
        """
        try:
            operating_income_per_share = (
                finance_dict['operating_income'] /
                finance_dict['share_outstanding']
            )
            finance_dict['operating_income_per_share'
                         ] = StatsProcessor.cal_non_percentage(
                             operating_income_per_share
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['operating_income_per_share'] = None
            # print(f"operating_income_per_share failed because of {str(e)}")

    @classmethod
    def cal_operating_cash_flow_per_share(cls, finance_dict):
        """
        計算每股營業現金流
        = (當期營業現金流) ÷（當期在外流通股數）
        """
        try:
            operating_cash_flow_per_share = (
                finance_dict["operating_cash_flow"] /
                finance_dict['share_outstanding']
            )
            finance_dict["operating_cash_flow_per_share"
                         ] = StatsProcessor.cal_non_percentage(
                             operating_cash_flow_per_share
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['operating_cash_flow_per_share'] = None
            # print(f'operating_cash_flow_per_share because of {str(e)}')

    @classmethod
    def fcf_per_share(cls, finance_dict):
        """
        計算每股自由現金流
        每股自由現金流 = (當期自由現金流) ÷（當期在外流通股數）
        """
        try:
            fcf_per_share = (
                finance_dict['fcf'] / finance_dict['share_outstanding']
            )
            finance_dict['fcf_per_share'] = StatsProcessor.cal_non_percentage(
                fcf_per_share
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['fcf_per_share'] = None
            # print(f"fcf_per_share failed because of {str(e)}")

# 盈利能力

    @classmethod
    def cal_roa(cls, finance_dict):
        """
        計算資產報酬率(ROA)
        ROA = [ 本期淨利 + 利息費用 × (1-有效稅率) ] ÷（資產總額）
        """
        try:
            roa = (
                finance_dict['net_income'] + finance_dict['interest'] +
                (1 * 0.1)    # 有效稅率需要改，這裡先設0.1
            ) / finance_dict['inventories']

            finance_dict["roa"] = StatsProcessor.cal_percentage(roa)
        except Exception as e:
            finance_dict["roa"] = None

    @classmethod
    def cal_roe(cls, finance_dict):
        """
        計算股東權益報酬率(ROE)
        ROE = (本期淨利) ÷（權益總額）
        """
        try:
            roe = (finance_dict['net_income'] / finance_dict['equity'])
            finance_dict['roe'] = StatsProcessor.cal_percentage(roe)
        except Exception as e:
            finance_dict['roe'] = None

    @classmethod
    def cal_gross_over_asset(cls, finance_dict):
        """
        計算營業毛利/總資產
        """
        try:
            gross_over_asset = (
                finance_dict['gross_profit'] / finance_dict['total_assets']
            )
            finance_dict['gross_over_asset'] = StatsProcessor.cal_percentage(
                gross_over_asset
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['gross_over_asset'] = None
            # print(f"營業毛利/總資產 failed because of {str(e)}")

    @classmethod
    def cal_roce(cls, finance_dict):
        """
        計算資本運用報酬率(ROCE):
        ROCE = (稅前淨利＋利息費用) / (資產總額－流動負債)
        """
        try:
            roce = (
                (
                    finance_dict['net_income_before_tax'] +
                    finance_dict['interest']
                ) / (
                    finance_dict['total_assets'] -
                    finance_dict['current_liabilities']
                )
            )
            finance_dict['roce'] = StatsProcessor.cal_percentage(roce)

        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['roce'] = None
            # print(f"ROCE failed because of {str(e)}")

    @classmethod
    def cal_gross_profit_marginal(cls, finance_dict):
        """
        計算營業毛利率(gross profit margin)
        營業毛利率 = 營業毛利 ÷ 營業收入
        """
        try:
            gross_profit_margin = (
                finance_dict['gross_profit'] / finance_dict['revenue']
            )
            finance_dict['gross_profit_margin'] = StatsProcessor.cal_percentage(
                gross_profit_margin
            )
        except Exception as e:
            finance_dict['gross_profit_margin'] = None
            # print(f"gross_profit_margin failed because of {str(e)}")

    @classmethod
    def cal_operation_profit_rate(cls, finance_dict):
        """
        計算營業利益率
        營業利益率 = ( 營業收入－營業成本－營業費用）÷ 營業收入
        """
        try:
            operation_profit_rate = (
                finance_dict['revenue'] - finance_dict['operating_cost'] -
                finance_dict['operating_expenses']
            ) / finance_dict['revenue']
            finance_dict["operation_profit_rate"
                         ] = StatsProcessor.cal_percentage(
                             operation_profit_rate
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["operation_profit_rate"] = None
            # print(f"operation_profit failed because of {str(e)}")

    @classmethod
    def cal_operating_cash_flow_profit_rate(cls, finance_dict):
        """
        計算營業現金流利潤率
        營業現金流利潤率 = 營業活動現金流 ÷ 營業收入
        """
        try:
            operating_cash_flow_profit_rate = (
                finance_dict["operating_cash_flow"] / finance_dict["revenue"]
            )
            finance_dict["operating_cash_flow_profit_rate"
                         ] = StatsProcessor.cal_percentage(
                             operating_cash_flow_profit_rate
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["operating_cash_flow_profit_rate"] = None
            # print(
            #     f"operating_cash_flow_profit_rate failed because of {str(e)}")


# 成長動能

    """
    前四個已經有了 revenue_YoY, gross_prof_YoY, operating_income_YoY, net_income_YoY:
    後四個在金流，還需要處理
    """
    # 營運指標

    @classmethod
    def cal_dso(cls, finance_dict):
        """
        計算應收帳款收現天數(DSO)
        DSO = 365 × (應收帳款平均餘額 ÷ 營業收入)
        """
        try:
            dso = (
                365 * (finance_dict['account_pay'] / finance_dict['revenue'])
            )
            finance_dict['dso'] = StatsProcessor.cal_non_percentage(
                dso, to_str=True, postfix="日"
            )
        except Exception as e:
            finance_dict['dso'] = None
            # print(f"Error calculating 應收帳款收現天數 because of {str(e)}")

    @classmethod
    def cal_account_receive_over_revenue(cls, finance_dict):
        """
        計算應收帳款佔營收比率
        = 應收帳款平均餘額 ÷ 營業收入
        """
        try:
            account_receive_over_revenue = (
                finance_dict['account_receive'] / finance_dict['revenue']
            )
            finance_dict["account_receive_over_revenue"
                         ] = StatsProcessor.cal_percentage(
                             account_receive_over_revenue
                         )
        except Exception as e:
            finance_dict["account_receive_over_revenue"] = None

    @classmethod
    def cal_dpo(cls, finance_dict):
        """
        計算應付帳款週轉天數
        DPO = 365天 ÷ (銷貨成本÷平均應付帳款)
        """
        try:
            dpo = (
                365 *
                (finance_dict['account_pay'] / finance_dict['operating_cost'])
            )
            finance_dict["dpo"] = StatsProcessor.cal_non_percentage(
                dpo, to_str=True, postfix="日"
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["dpo"] = None
            # print(f"應付帳款週轉天數 failed because of {str(e)}")

    @classmethod
    def cal_inventories_cycle_ratio(cls, finance_dict):
        """
        計算存貨周轉率
        = 銷貨成本 ÷ 存貨
        """
        try:
            inventories_cycle_ratio = (
                finance_dict['operating_cost'] / finance_dict['inventories']
            )

            finance_dict["inventories_cycle_ratio"
                         ] = StatsProcessor.cal_percentage(
                             inventories_cycle_ratio
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["inventories_cycle_ratio"] = None
            # print(f"Error calculating 存貨周轉率 because of {str(e)}")

    @classmethod
    def cal_dio(cls, finance_dict):
        """
        計算 存貨週轉天數 or 平均售貨天數
        DIO = 365天 * (存貨 ÷ 銷貨成本)
        MUDA MUDA MUDA !!!
        """
        try:
            dio = 365 * (
                finance_dict["inventories"] / finance_dict["operating_cost"]
            )
            finance_dict["dio"] = StatsProcessor.cal_non_percentage(
                dio, to_str=True, postfix="日"
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["dio"] = None
            # print(f"Error calculating 存貨週轉天數 because of {str(e)}")

    @classmethod
    def cal_inventories_revenue_ratio(cls, finance_dict):
        """
        計算存貨佔營收比率
        存貨佔營收比= 存貨 ÷ 營業收入
        """
        try:
            inventories_revenue_ratio = (
                finance_dict['inventories'] / finance_dict['revenue']
            )

            finance_dict["inventories_revenue_ratio"
                         ] = StatsProcessor.cal_percentage(
                             inventories_revenue_ratio
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["inventories_revenue_ratio"] = None
            # print(f"Error calculating 存貨佔營收比率 because of {str(e)}")

    @classmethod
    def cal_cash_of_conversion_cycle(cls, finance_dict):
        """
        計算現金循環週期
        存貨週轉天數 + 應收帳款週轉天數 - 應付帳款週轉天數
        """
        try:
            cash_of_conversion_cycle = (
                finance_dict["dio"] + finance_dict["dso"] - finance_dict['dpo']
            )
            finance_dict["cash_of_conversion_cycle"
                         ] = StatsProcessor.cal_non_percentage(
                             cash_of_conversion_cycle, to_str=True, postfix="日"
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["cash_of_conversion_cycle"] = None

    @classmethod
    def cal_asset_turnover(cls, finance_dict):
        """
        計算資產周轉率
        營業收入 ÷ 資產總額
        """
        try:
            asset_turnover = (
                finance_dict["revenue"] / finance_dict["inventories"]
            )
            finance_dict["asset_turnover"] = StatsProcessor.cal_percentage(
                asset_turnover
            )
        except Exception as e:
            finance_dict["asset_turnover"] = None

    @classmethod
    def cal_application_turnover(cls, finance_dict):
        """
        不動產、廠房及設備週轉率
        營業收入 ÷ 不動產、廠房與設備平均餘額
        """
        try:
            applcation_turnover = (
                finance_dict['revenue'] / finance_dict["application"]
            )
            finance_dict['application_turnover'
                         ] = StatsProcessor.cal_percentage(applcation_turnover)

        except Exception as e:
            finance_dict['application_turnover'] = None

    @classmethod
    def cal_current_ratio(cls, finance_dict):
        """
        計算流動比率 = 流動資產 / 流動負債
        """
        try:
            current_ratio = (
                finance_dict['current_assets'] /
                finance_dict['current_liabilities']
            )
            finance_dict['current_ratio'] = StatsProcessor.cal_percentage(
                current_ratio
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['current_ratio'] = None
            # print(f"Error calculating current ratio: {e}")

    @classmethod
    def cal_quick_ratio(cls, finance_dict):
        """
        速動比率
        (流動資產 - 存貨) / 流動負債
        """
        try:
            quick_ratio = (
                finance_dict['current_assets'] - finance_dict['inventories']
            ) / finance_dict['current_liabilities']
            finance_dict['quick_ratio'] = StatsProcessor.cal_percentage(
                quick_ratio
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['quick_ratio'] = None
            # print(f"Error calculating quick ratio: {e}")

    @classmethod
    def cal_debt_to_equity_ratio(cls, finance_dict):
        """
        # 負債權益比率 = 總負債 / 股東權益
        """
        try:
            debt_to_equity_ratio = finance_dict['total_liabilities'
                                                ] / finance_dict['equity']
            finance_dict['debt_to_equity_ratio'
                         ] = StatsProcessor.cal_percentage(
                             debt_to_equity_ratio
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['debt_to_equity_ratio'] = None
            # print(f"Error calculating debt to equity ratio: {e}")

    @classmethod
    def cal_net_debt_to_equity_ratio(cls, finance_dict):
        """
        # 淨負債權益比率 = (總負債 - 現金及約當現金) / 股東權益
        """
        try:
            net_debt_to_equity_ratio = (
                finance_dict['total_liabilities'] -
                finance_dict['cash_and_cash_equivalents']
            ) / finance_dict['equity']
            finance_dict['net_debt_to_equity_ratio'
                         ] = StatsProcessor.cal_percentage(
                             net_debt_to_equity_ratio
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['net_debt_to_equity_ratio'] = None
            # print(f"Error calculating net debt to equity ratio: {e}")

    @classmethod
    def cal_interest_coverage_ratio(cls, finance_dict):
        """
        # 利息保障倍數 = EBIT / 利息費用
        """
        try:
            interest_coverage_ratio = finance_dict['EBIT'] / finance_dict[
                'interest_expense']
            finance_dict['interest_coverage_ratio'
                         ] = StatsProcessor.cal_non_percentage(
                             interest_coverage_ratio, to_str=True, postfix="倍"
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['interest_coverage_ratio'] = None
            # print(f"Error calculating interest coverage ratio: {e}")

    @classmethod
    def cal_debt_to_operating_cash_flow(cls, finance_dict):
        """
        有息負債 / 營業活動現金流
        """
        try:
            debt_to_operating_cash_flow = finance_dict[
                'interest_bearing_debt'] / finance_dict['operating_cash_flow']
            finance_dict['debt_to_operating_cash_flow'
                         ] = StatsProcessor.cal_percentage(
                             debt_to_operating_cash_flow
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['debt_to_operating_cash_flow'] = None
            # print(f"Error calculating debt to operating cash flow: {e}")

    @classmethod
    def cal_debt_to_free_cash_flow(cls, finance_dict):
        """
        # 有息負債 / 自由現金流
        """
        try:
            debt_to_free_cash_flow = finance_dict['interest_bearing_debt'
                                                  ] / finance_dict['fcf']
            finance_dict['debt_to_free_cash_flow'
                         ] = StatsProcessor.cal_percentage(
                             debt_to_free_cash_flow
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['debt_to_free_cash_flow'] = None
            # print(f"Error calculating debt to free cash flow: {e}")

    @classmethod
    def cal_cash_flow_ratio(cls, finance_dict):
        """
        # 現金流量比率 = 營業活動現金流 / 流動負債
        """
        try:
            cash_flow_ratio = finance_dict[
                'operating_cash_flow'] / finance_dict['current_liabilities']
            finance_dict['cash_flow_ratio'] = StatsProcessor.cal_percentage(
                cash_flow_ratio
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['cash_flow_ratio'] = None
            # print(f"Error calculating cash flow ratio: {e}")


class AgentFinanceOverviewFetcher(FinanceOverviewFetcher):
    """
    目前僅適用美股
    """

    def __init__(self, ticker, db_client):
        self.ticker = ticker
        self.timezone = pytz.timezone("Asia/Taipei")
        self.tw_company_list = StatsProcessor.load_json("company_list/tw.json")
        self.us_company_list = StatsProcessor.load_json("company_list/us_TradingView_list.json")

        if (ticker not in self.us_company_list.keys()):
            # 目前只支援美股
            raise NoCompanyError("class \"AgentFinanceOverviewFetcher\" only supports US company now")

        db_name = "company" if self.ticker in self.tw_company_list else "company_us"
        self.db = db_client[db_name]

        daily_collection_name_map = {
            "company": "twse_daily_share_price",
            'company_us': "us_technical"
        }

        seasonal_collection_name_map = {
            'company': "twse_seasonal_report",
            "company_us": "us_fundamentals"
        }

        self.daily_collection_name = daily_collection_name_map.get(db_name)
        self.seasonal_collection_name = seasonal_collection_name_map.get(
            db_name
        )

        assert self.daily_collection_name != "unknown", f"請確認 {ticker} 是否是 {','.join(list(daily_collection_name_map.values()))}"
        assert self.seasonal_collection_name != "unknown", f"請確認 {ticker} 是否是 {','.join(list(seasonal_collection_name_map.values()))}"
        self.daily_collection = db_client[db_name][self.daily_collection_name]
        self.seasonal_collection = db_client[db_name][
            self.seasonal_collection_name]

    def query_data(self, date=None):
        query_set = self._prepare_query(date)
        if (date is None):
            str_date = datetime.today().strftime("%Y-%m-%d")
        if (isinstance(date, datetime)):
            str_date = date.strftime("%Y-%m-%d")
        elif (isinstance(date, str)):
            str_date = date

        daily_query, daily_projection, daily_sorting = query_set['daily']
        seasonal_query, seasonal_projection, seasonal_sorting = query_set[
            'seasonal']

        daily_datas = self.daily_collection.find(daily_query, daily_projection
                                                 ).sort(daily_sorting)
        seasonal_datas = self.seasonal_collection.find(
            seasonal_query, seasonal_projection
        ).sort(seasonal_sorting)

        daily_data = [data for data in daily_datas]
        seasonal_data = [
            {
                **data['balance_sheet'],
                **data['income_statement'],
                **data['cash_flow']
            } for data in seasonal_datas
        ]

        seasonal_df = pd.DataFrame(seasonal_data)
        TTM_data = seasonal_df.mean()

        return_dict = {
            "date": str_date,
            "ticker": self.ticker,
            "company_name": self._get_company_name(),
            "category": self._get_category(),
            "market_open": self._is_us_market_open(),
            "latest_close":
            self._get_latest_close(daily_data),
            "close_offset":
            self._get_latest_offset(daily_data).get("value", "0.0"),
            "close_offset_percentage":
            self._get_latest_offset(daily_data).get("percentage", "0.0%"),
            "latest_volume":
            self._get_latest_volume(daily_data),
            "average_volume":
            self._get_latest_average_volume(daily_data, 30),
            "market_capitalzation":
            self._get_market_capitalization(daily_data, seasonal_data),
            "P_E":
            self._get_PE(daily_data, TTM_data),
            "P_S":
            self._get_PS(daily_data, TTM_data, seasonal_data),
        }

        return return_dict

    def _prepare_query(self, date=None):
        if (date is None):
            date = datetime.today()
        elif (isinstance(date, str)):
            date = datetime.strptime(date, "%Y-%m-%d")

        daily_query = {"ticker": self.ticker, "date": {"$lte": date}}

        seasons = []
        year = date.year
        season = (date.month - 1) // 3 + 1

        for _ in range(4):
            seasons.append({"year": year, "season": season})
            year = year if season != 1 else year - 1
            season = season - 1 if season != 1 else 4

        seasonal_query = {"ticker": self.ticker, "$or": seasons}

        daily_projection = {"_id": 0}
        seasonal_projection = {"_id": 0}

        daily_sorting = {"date": ASCENDING}
        seasonal_sorting = {"year": ASCENDING, "season": ASCENDING}

        query_set = {
            "daily": [daily_query, daily_projection, daily_sorting],
            "seasonal": [seasonal_query, seasonal_projection, seasonal_sorting]
        }

        return query_set

    def _get_latest_close(self, daily_data):
        latest_close = daily_data[-1]['close']
        latest_close = str(round(latest_close, 2))
        return latest_close

    def _get_latest_offset(self, daily_data):
        latest_close = daily_data[-1]['close']
        previous = daily_data[-2]['close']

        offset = latest_close - previous
        percentage = offset / previous

        offset = str(np.round(offset, 2))

        return {'value': offset, "percentage": f"{percentage:.2f}%"}

    def _get_latest_volume(self, daily_data):
        try:
            volume = round(daily_data[-1]['volume'], 2)
            return str(volume)
        except Exception as e:
            return "Error"

    def _get_latest_average_volume(self, daily_data, length=30):
        """
        計算平均成交量
        """
        try:
            daily_data = pd.DataFrame(daily_data).tail(length)

            value = daily_data['volume'].mean().item()
            value = round(value, 2)

            return str(value)

        except Exception as e:
            return "Error"

    def _get_market_capitalization(self, daily_data, seasonal_data):
        """
        計算市值
        市值 = 收盤價 * 流通股數
        """
        try:
            latest_close = float(self._get_latest_close(daily_data))
            latest_common_share = seasonal_data[-1][
                'Common Stock']

            percentage = round((latest_close * latest_common_share), 2)

            return str(percentage)

        except Exception as e:
            return "Error"

    def _get_PE(self, daily_data, TTM_data):
        """
        本益比
        本益比 = 收盤價 / 稀釋EPS(TTM)
        """
        try:
            latest_close = float(self._get_latest_close(daily_data))
            TTM_EPS = TTM_data['Diluted EPS']

            value = round((latest_close / TTM_EPS), 2)

            return str(value)

        except Exception as e:
            return "Error"

    def _get_PS(self, daily_data, TTM_data, seasonal_data):
        """
        股價營收比: 
        本益比 = 市值 / 總營收(TTM)
        """
        try:
            market_capitalzation = self._get_market_capitalization(daily_data, seasonal_data)
            market_capitalzation = float(market_capitalzation)
            TTM_revenue = TTM_data['Total Revenue']

            value = round((market_capitalzation / TTM_revenue), 2)

            return str(value)
        except Exception as e:
            return "Error"

    def _is_us_market_open(self):
        """
        判斷目前美股是否開市
        """
        taiwan_dt = datetime.now(pytz.timezone('Asia/Taipei'))

        # 轉成美東時間（會自動處理夏令時間）
        eastern = pytz.timezone('US/Eastern')
        us_dt = taiwan_dt.astimezone(eastern)

        # 假日判斷
        us_holidays = holidays.NYSE(years=us_dt.year)
        if us_dt.date() in us_holidays:
            return False

        # 週末
        if us_dt.weekday() >= 5:  # 5 = Saturday, 6 = Sunday
            return False

        # 判斷是否在開盤時間
        market_open = us_dt.replace(hour=9, minute=30, second=0, microsecond=0)
        market_close = us_dt.replace(hour=16, minute=0, second=0, microsecond=0)
        
        return market_open <= us_dt <= market_close
    
    def _get_category(self):
        """
        判斷公司類別，透過檔案
        """
        category = self.us_company_list.get(self.ticker,{}).get('en_industry')

        if (category):
            return category

        else:
            raise ValueError("No company")
    
    def _get_company_name(self):
        company_name = self.us_company_list.get(self.ticker,{}).get('name')

        if (company_name):
            return company_name

        else:
            raise ValueError("No company")


from .base import BaseDailyTransformer
from .utils import TechProcessor

import inspect
import pandas as pd
from datetime import datetime
from typing import List, Optional, Union
from collections import defaultdict

class DailyTechTransformer(BaseDailyTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

        self.basic_indexes = [
            'SMA5', 'SMA20', 'SMA60', 'EMA5', 'EMA20', 'EMA40', 'EMA12',
            'EMA26', 'RSI7', 'RSI14', 'RSI21', 'MACD', 'Signal Line',
            'Middle Band', 'Upper Band', 'Lower Band', '%b', 'ATR', 'BBW',
            'EMA Cycle', 'EMA Cycle Instructions', 'Day Trading Signal'
        ]

        self.fetched_data = None
        self.daily_index = None
        self.weekly_index = None
        self.monthly_index = None
        self.quarterly_index = None
        self.yearly_index = None
    
    def process_transform(self, fetched_data):
        self.fetched_data = pd.DataFrame(fetched_data)

        self.daily_index = TechProcessor.cal_basic_index(
            self.fetched_data
        )
    
    def _check_daily_index(self):
        """
        檢查 daily_index 是否已經計算。
        如果 daily_index 為 None，則藉由 inspect 模組取得呼叫該函式的 caller，
        並拋出錯誤訊息，提示先呼叫 process_transform(fetched_data)
        再呼叫對應的函數（例如 get_weekly()）。
        """
        if self.daily_index is None:
            # 取得呼叫此 _check_daily_index 的函數名稱
            caller_name = inspect.stack()[1].function
            raise ValueError(f"Please call process_transform(fetched_data) before accessing {caller_name}()")

    def _get_resampled(self, period: str):
        """
        根據指定時間間隔對 daily_index 進行重採樣。
        """
        self._check_daily_index()  # 如果 daily_index 尚未計算，會拋出錯誤
        return TechProcessor.resample(
            self.daily_index,
            period=period,
            technical_indicators=self.basic_indexes
        )
    
    def get_daily(self):
        """
        回傳日線
        """
        self._check_daily_index() 
        return self.daily_index

    def get_weekly(self):
        """
        回傳週線資料。
        """
        return self._get_resampled('W')

    def get_monthly(self):
        """
        回傳月線資料。
        """
        return self._get_resampled('ME')

    def get_quarterly(self):
        """
        回傳季線資料。
        """
        return self._get_resampled('QE')
    
    def get_yearly(self):
        """
        回傳年線資料
        """
        return self._get_resampled("YE")
    
class BatchTechTransformer(BaseDailyTransformer):
    def __init__(self, strftime:str="%Y-%m-%d"):
        self.date_format = strftime

    def process_transform(
        self,
        fetched_datas: List[dict],
    ) -> dict[str, dict[Union[str, datetime], float]]:
        
        result = defaultdict(lambda: defaultdict(dict))

        for record in fetched_datas:
            ticker = record.get("ticker")
            date = record.get("date")
            if isinstance(date, str):
                date = datetime.fromisoformat(date)
            date_str = date.strftime(self.date_format)

            result[ticker][date_str] = {
                "open": record.get("open"),
                "high": record.get("high"),
                "low": record.get("low"),
                "close": record.get("close"),
                "volume": record.get("volume"),
            } 

        return dict(result)


from datetime import datetime
from pytz import timezone

class StatsDateTime():

    def __init__(self, date, year, month, day, season):
        self.date = date
        self.year = year
        self.month = month
        self.day = day
        self.season = season
    
    @classmethod
    def get_today(cls):
        today = datetime.today()
        this_year = today.year
        this_month = today.month
        this_day = today.day
        this_season = ((this_month - 1) // 3) + 1

        return StatsDateTime(today, this_year, this_month, this_day,
                             this_season)

    @classmethod
    def get_latest_time(cls, ticker, collection):
        return collection.find_one(
            { "ticker" : ticker }, 
            { "_id": 0, "last_update_time": 1 }
        )

    @classmethod
    def transform_date(cls, date, tz:str = None):
        if (isinstance(date, str)):
            try:
                date = datetime.strptime(date, "%Y-%m-%d")
                if (tz):
                    date = date.astimezone(tz = timezone(tz))
                return date 
            except Exception as e:
                print(e)
                raise ValueError(
                    "date string format imcompatible, should be \"YYYY-MM-DD\""
                )
        elif (isinstance(date, datetime)):
            if (tz):
                return date.astimezone(tz = timezone(tz))
            else:
                return date
        else:
            raise ValueError(
                "date type imcompatible, should be string(\"YYYY-MM-DD\") or datetime.datetime"
            )


from .base import BaseDailyTechDBExtractor
from datetime import datetime
from pymongo import ASCENDING, DESCENDING
from neurostats_API.utils import NotSupportedError

class AsyncTWSEChipDBExtractor(BaseDailyTechDBExtractor):

    def __init__(self, ticker, client, fetch_type = 'I'):

        """
        擷取台股每日籌碼資料。

        fetch_type:
            - 'I' (INSTITUTION): 法人買賣
            - 'M' (MARGIN): 融資融券餘額變化
            - 'S' (SECURITY_LENDING): 借券債券

        """
        super().__init__(ticker, client)

        column_name_map = {
            'tw': {
                "I" : "institution_trading",
                "M" : "margin_trading",
                "S" : "security_lending"
            },
        }
        self.target_column = column_name_map[self.zone][fetch_type]

    def _get_collection_name(self):
        self.collection_name_map = {"tw": "twse_chip"}
        collection_name = self.collection_name_map.get(self.zone, None)
        return collection_name
    
    def _prepare_query(self, start_date=None, end_date=None, get_latest = False):
        query, projection, sort = super()._prepare_query(start_date, end_date, get_latest)
        query.update(
            {
                self.target_column: {"$exists" : True}
            }
        )
        projection.update(
            {
                'date': 1,
                self.target_column: 1
            }
        )

        return query, projection, sort



class NoCompanyError(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class NoDataError(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class NotSupportedError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


from .balance_sheet import (
    TWSEBalanceSheetTransformer,
    USBalanceSheetTransformer
)

from .cash_flow import (
    TWSECashFlowTransformer,
    USCashFlowTransformer
)

from .daily_tech import(
    DailyTechTransformer,
    BatchTechTransformer
)

from .daily_chip import(
    TWSEChipTransformer,
    US_F13_Transformer,
    US_CelebrityTransformer
)

from .month_revenue import(
    TWSEMonthlyRevenueTransformer
)

from .profit_lose import (
    TWSEProfitLoseTransformer,
    USProfitLoseTransformer
)

from .tej import (
    TEJChipTransformer,
    TEJFinanceStatementTransformer,
    TEJTechTransformer
)

from .value import (
    TWSEAnnualValueTransformer,
    TWSEHistoryValueTransformer
)

from .finance_overview import (
    AgentOverviewTransformer,
    FinanceOverviewTransformer
)


from .base import StatsFetcher, StatsDateTime
import json
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor, YoY_Calculator
import importlib.resources as pkg_resources
import yaml
import traceback
import logging


class MonthRevenueFetcher(StatsFetcher):
    """
        iFa.ai: 財務分析 -> 每月營收
    """

    def __init__(self, ticker, db_client, logger = None):
        super().__init__(ticker, db_client)
        self.logger = logger or logging.getLogger(__name__)

    def _prepare_query(self, target_year, target_month):
        pipeline = super().prepare_query()

        pipeline.append(
            {
                "$project": {
                    "_id": 0,
                    "ticker": 1,
                    "company_name": 1,
                    "monthly_data": {
                        "$sortArray": {
                            "input": "$monthly_data",
                            "sortBy": {
                                "year": -1,
                                "month": -1
                            }
                        }
                    },
                }
            }
        )

        return pipeline

    def collect_data(self, target_year, target_month):
        pipeline = self._prepare_query(target_year, target_month)
        fetched_data = self.collection.aggregate(pipeline)
        fetched_data = fetched_data.to_list()

        return fetched_data

    def query_data(self):
        target_year, target_month = self._get_target_year_and_month()

        # Query data
        fetched_data = self.collect_data(target_year, target_month)

        try:
            return self._process_data(fetched_data[-1])
        except Exception:
            recent_date = []
            for _ in range(12):
                recent_date.append(f"{target_year}/{target_month}")
                target_year, target_month = (
                    target_year - 1, 12
                ) if target_month == 1 else (target_year, target_month - 1)

            # logging.warning(f"{self.ticker}: No monthly revenue data in TWSE mongoDB", exc_info=True)
            return self._get_empty_structure(target_year, target_month)

    def _process_data(self, fetched_data):

        monthly_data = fetched_data.get('monthly_data', [])
        if not monthly_data:
            raise ValueError("monthly_data is empty or missing")

        for data in monthly_data:
            for key, value in data.items():
                if "YoY" in key:
                    data[key] = StatsProcessor.cal_percentage(value)
                elif "ratio" in key or "percentage" in key:
                    data[key] = StatsProcessor.cal_non_percentage(value, to_str=True, postfix="%")
                elif key not in ('year', 'month'):
                    data[key] = StatsProcessor.cal_non_percentage(value, postfix="千元")

        monthly_df = pd.DataFrame(monthly_data)
        target_month = monthly_data[0]['month']
        target_month_df = monthly_df[monthly_df['month'] == target_month]
        annual_month_df = monthly_df[monthly_df['month'] == 12]

        
        month_revenue_df = monthly_df.pivot(
            index='month', columns='year', values='revenue'
        )
        month_revenue_df = month_revenue_df.sort_index(ascending=False)

        grand_total_df = target_month_df.pivot(
            index='month', columns='year', values='grand_total'
        )
        grand_total_df.rename(
            index={target_month: f"grand_total"}, inplace=True
        )
        month_revenue_df = pd.concat([grand_total_df, month_revenue_df], axis=0)

        annual_total_df = annual_month_df.pivot(
            index='month', columns='year', values='grand_total'
        )

        fetched_data.update(
            {
                "month_revenue": month_revenue_df[sorted(month_revenue_df.columns, reverse=True)],
                "this_month_revenue_over_years": target_month_df.set_index("year")[[
                    "revenue", "revenue_increment_ratio", "YoY_1", "YoY_3", "YoY_5", "YoY_10"
                ]].T,
                "grand_total_over_years": target_month_df.set_index("year")[[
                    "grand_total", "grand_total_increment_ratio", "grand_total_YoY_1", "grand_total_YoY_3", "grand_total_YoY_5", "grand_total_YoY_10"
                ]].T,
                "recent_month_revenue": self._get_recent_growth(monthly_data, grand_total_dict=annual_total_df.to_dict(), interval=12)
            }
        )

        fetched_data.pop("monthly_data")
        return fetched_data

    def _get_recent_growth(self, monthly_data, grand_total_dict, interval=12):
        last_month_data = monthly_data[1:interval + 1] + [{}] * max(0, interval - len(monthly_data) + 1)

        MoMs = [
            YoY_Calculator.cal_growth(this.get('revenue'), last.get('revenue'), delta = 1)
            for this, last in zip(monthly_data[:interval], last_month_data[:interval])
        ]

        def safe_accum_yoy(data):
            try:
                year = data['year'] - 1
                total = grand_total_dict[year][12]
                grand_total = data.get('grand_total')
                return f"{round(((grand_total - total) / total) * 100, 2)}%"
            except Exception:
                self.logger.debug(f"accum_YoY calc failed for year={data.get('year')} / ticker={self.ticker}", exc_info=True)
                return None

        recent_month_data = {
            "date": [f"{d.get('year', 0)}/{d.get('month', 0)}" for d in monthly_data[:interval]],
            "revenue": [d.get('revenue') for d in monthly_data[:interval]],
            "MoM": [f"{(m * 100):.2f}%" if isinstance(m, float) else None for m in MoMs],
            "YoY": [d.get('revenue_increment_ratio') for d in monthly_data[:interval]],
            "total_YoY": [d.get('grand_total_increment_ratio') for d in monthly_data[:interval]],
            # accum_YoY
            # accum_YoY 為 Davis提出的定義
            # 2024/6的累計YoY(accum_YoY) 為 2024累計到6月為止的總營收/2023年度總營收
            "accum_YoY": [safe_accum_yoy(d) for d in monthly_data[:interval]]
        }

        df = pd.DataFrame(recent_month_data)
        return df[df['date'] != "0/0"].set_index('date').T


    def _get_empty_structure(self, target_year, target_month):
        """
        Exception 發生時回傳
        """
        recent_date = [f"{target_year}/{target_month}"]
        for _ in range(11):
            target_year, target_month = (target_year - 1, 12) if target_month == 1 else (target_year, target_month - 1)
            recent_date.append(f"{target_year}/{target_month}")

        def empty_df(index, columns):
            return pd.DataFrame(index=index, columns=columns)

        return {
            "month_revenue": empty_df(
                index=pd.Index(['grand_total', 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1], dtype='object', name='month'),
                columns=pd.Index([f"{target_year - i}" for i in range(10)], dtype=object, name='year')
            ),
            "this_month_revenue_over_years": empty_df(
                index=pd.Index(['revenue', 'revenue_increment_ratio', 'YoY_1', 'YoY_3', 'YoY_5', 'YoY_10'], dtype='object'),
                columns=pd.Index([f"{target_year - i}" for i in range(10)], dtype='int64', name='year')
            ),
            "grand_total_over_years": empty_df(
                index=pd.Index(['grand_total', 'grand_total_increment_ratio', 'grand_total_YoY_1', 'grand_total_YoY_3', 'grand_total_YoY_5', 'grand_total_YoY_10'], dtype='object'),
                columns=pd.Index([f"{target_year - i}" for i in range(10)], dtype='int64', name='year')
            ),
            "recent_month_revenue": empty_df(
                index=pd.Index(['revenue', 'MoM', 'YoY', 'total_YoY', 'accum_YoY'], dtype='object'),
                columns=pd.Index([], dtype = 'object', name = 'date')
            )
        }
    
    def _get_target_year_and_month(self):
        try:
            latest_time = StatsDateTime.get_latest_time(self.ticker, self.collection)['last_update_time']
            return latest_time['monthly_data']['latest_year'], latest_time['monthly_data']['latest_month']
        except Exception:
            today = StatsDateTime.get_today()
            return today.year, today.month


from ..base import BaseDBExtractor
from datetime import datetime
import json
import pandas as pd
from pymongo import ASCENDING, DESCENDING
from neurostats_API.async_mode.db import TWSEDBClient, USDBClient
from neurostats_API.utils import StatsDateTime, StatsProcessor, NotSupportedError
import yaml


class AsyncBaseSeasonalDBExtractor(BaseDBExtractor):

    def __init__(self, ticker, client):
        super().__init__(ticker, client)

        self.collection_name_map = self._get_collection_map()
        self.collection_name = self._get_collection_name()
        if (self.collection_name is None):
            raise NotSupportedError(
                f"{self.__class__.__name__} only supports {list(self.collection_name_map.keys())}, got {self.zone} with ticker = \"{self.ticker}\""
            )
        self.collection = self.db_client.get_collection(self.collection_name)

        self.column_name_map = None

    async def query_data(
        self,
        start_date=None,
        end_date=None,
        get_latest=False    # 決定是否只取得最新一筆資料
    ):
        if (start_date is None):
            start_date = "1991-01-01"
        if (end_date is None):
            end_date = self._get_today()

        start_year, start_season = self._get_year_and_season(date=start_date)
        end_year, end_season = self._get_year_and_season(date=end_date)

        query, projection, sort = self._prepare_query(
            start_year, start_season, end_year, end_season, get_latest
        )
        cursor = self.collection.find(query, projection).sort(sort)

        if (get_latest):
            cursor = cursor.limit(1)

        fetched_data = [data async for data in cursor]

        return fetched_data

    def _get_collection_name(self):

        return self.collection_name_map.get(self.zone, None)

    def _get_collection_map(self):
        return {"tw": "twse_seasonal_report", "us": "us_fundamentals"}

    def _prepare_query(
        self,
        start_year=None,
        start_season=None,
        end_year=None,
        end_season=None,
        get_latest=False
    ):
        chart_name = self.column_name_map.get(self.get_zone(), None)

        query = {
            "ticker": self.ticker,
            chart_name: {"$exists": True}
        }

        query = self._update_query_with_year_season(
            query, start_year, start_season, end_year, end_season
        )

        projection = {
            "_id": 0,
            "ticker": 1,
            "year": 1,
            "season": 1,
            chart_name: 1,
        }

        if (get_latest):
            sort = [("year", DESCENDING), ("season", DESCENDING)]
        else:
            sort = [("year", ASCENDING), ("season", ASCENDING)]

        return query, projection, sort

    def _get_year_and_season(self, date):
        if (isinstance(date, str)):
            date = datetime.strptime(date, "%Y-%m-%d")

        year = date.year
        month = date.month

        season = (month - 1) // 3 + 1

        return year, season

    def _update_query_with_year_season(
        self, query, start_year, start_season, end_year, end_season
    ):
        if all(v is not None for v in [start_year, start_season]):
            # 大於start_year條件
            query.update(
                {
                    "$or": [
                        {
                            "year": {
                                "$gt": start_year
                            }
                        },
                        {
                            "year": start_year,
                            "season": {
                                "$gte": start_season
                            }
                        },
                    ]
                }
            )

        if all(v is not None for v in [end_year, end_season]):
            # 小於end_year條件
            query.update(
                {
                    "$and":
                    query.get("$and", []) + [
                        {
                            "$or": [
                                {
                                    "year": {
                                        "$lt": end_year
                                    }
                                },
                                {
                                    "year": end_year,
                                    "season": {
                                        "$lte": end_season
                                    }
                                },
                            ]
                        }
                    ]
                }
            )

        return query



from .balance_sheet import AsyncBalanceSheetFetcher
from .cash_flow import AsyncCashFlowFetcher
from .finance_overview import AsyncFinanceOverviewFetcher, AsyncAgentOverviewFetcher
from .tech import AsyncTechFetcher, AsyncBatchTechFetcher
from .twse_institution import AsyncTWSEInstitutionFetcher
from .twse_margin import AsyncTWSEMarginFetcher
from .month_revenue import AsyncMonthlyRevenueFetcher
from .profit_lose import AsyncProfitLoseFetcher
from .value import AsyncTWSEStatsValueFetcher
from .tej import AsyncTEJChipFetcher, AsyncTEJDailyTechFetcher, AsyncTEJSeasonalFetcher
from .us_chip import AsyncUSChipF13Fetcher, AsyncUSCelebrityFetcher



from .base import BaseTEJTransformer
from neurostats_API.utils import StatsProcessor, NoCompanyError, NoDataError
import pandas as pd


class TEJFinanceStatementTransformer(BaseTEJTransformer):

    def __init__(self, ticker, company_name, zone, target_column="Q"):
        super().__init__(ticker, company_name, zone)
        self.target_column = target_column

        index_files = [
            "tej_db/tej_db_index.yaml", "tej_db/tej_db_thousand_index.yaml",
            "tej_db/tej_db_percent_index.yaml", "tej_db/tej_db_skip_index.yaml"
        ]

        self.index_dict, self.thousand_dict, self.percent_dict, skip_index_dict = [
            StatsProcessor.load_yaml(file) for file in index_files
        ]

        self.check_index = set(self.index_dict.get("TWN/AINVFQ1", []))
        self.skip_index = set(
            self.percent_dict.get("TWN/AINVFQ1", []) + skip_index_dict.get("TWN/AINVFQ1", [])
        )
        self.thousand_index_list = list(
            self.thousand_dict.get("TWN/AINVFQ1", [])
        )
        self.percent_index_list = list(self.percent_dict.get("TWN/AINVFQ1", []))

    def process_transform(
        self,
        fetched_data,
        start_date,
        fetch_mode='QoQ',
        use_cal=True,
        indexes=None
    ):
        if (not fetched_data):
            raise NoDataError(f"No data found in collection: TEJ_finance_statement, ticker={self.ticker}")

        target_season = fetched_data[-1]['season']
        start_col = self._find_start_season(fetched_data, start_date)
        fetched_data = self._process_data_to_tej_format(
            fetched_data, target_key=self.target_column
        )

        if (indexes):
            fetched_data = pd.DataFrame.from_dict(fetched_data)
            fetched_data = fetched_data.loc[indexes, :]
            fetched_data = fetched_data.to_dict()

        if (fetch_mode == 'QoQ'):
            data = self._get_QoQ_data(data=fetched_data, use_cal=use_cal)
            data = data.iloc[:, start_col:]
        else:
            data = self._get_YoY_data(
                data=fetched_data, target_season=target_season, use_cal=use_cal
            )

            data = data.iloc[:, start_col:]
            target_season_col = data.columns.str.endswith(f"{target_season}")
            data = data.loc[:, target_season_col]

        return data

    def _find_start_season(self, fetched_data, start_date):
        """
        找出開始的年或季度所在的index
        """
        start_year = start_date.year
        start_season = (start_date.month - 1 // 3) + 1

        start_col = 0
        for data in fetched_data:
            if ((start_year, start_season) < (data['year'], data['season'])):
                break
            else:
                start_col += 1

        return start_col



from .base import BaseBalanceSheetTransformer
from neurostats_API.utils import StatsProcessor


class USBalanceSheetTransformer(BaseBalanceSheetTransformer):

    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

        self.data_df = None
        self.return_keys = [
            'balance_sheet',
            'balance_sheet_YoY'
        ]

    def process_transform(self, fetched_data):
        if (not fetched_data):
            return self._get_empty_structure()

        return_dict = {"ticker": self.ticker, "company_name": self.company_name}

        # QoQ表格
        self.data_df = self._process_us_format(fetched_data)

        # YoY表格
        target_season = fetched_data[0]['season']
        total_table_YoY = self._slice_target_season(
            self.data_df.T, target_season
        )

        return_dict.update(
            {
                'balance_sheet': self.data_df,
                "balance_sheet_YoY": total_table_YoY
            }
        )

        return return_dict



from .base import StatsFetcher, StatsDateTime
import json
import numpy as np
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor
import importlib.resources as pkg_resources
import yaml

class CashFlowFetcher(StatsFetcher):
    def __init__(self, ticker, db_client):
        super().__init__(ticker, db_client)

        self.cash_flow_dict = StatsProcessor.load_yaml(
            "twse/cash_flow_percentage.yaml"
        )  # 計算子表格用

        self.process_function_map = {
            "twse_stats": self.process_data_twse,
            "us_stats": self.process_data_us
        }

        self.return_keys = ['cash_flow', 'CASHO', 'CASHI', 'CASHF', 'cash_flow_all', 'cash_flow_YoY']
    
    def prepare_query(self):
        pipeline = super().prepare_query()

        name_map = {
            "twse_stats": "cash_flow",
            "us_stats": "cash_flow"
        }


        chart_name = name_map.get(self.collection_name, "cash_flow")

        append_pipeline = [
        {
            "$project": {
                "_id": 0,
                "ticker": 1,
                "company_name": 1,
                "seasonal_data": {
                    "$map": {
                        "input": {"$ifNull": ["$seasonal_data", []]},
                        "as": "season",
                        "in": {
                            "year": "$$season.year",
                            "season": "$$season.season",
                            "data": {"$ifNull": [f"$$season.{chart_name}", []]}
                        }
                    }
                }
            }
        }
    ]

        pipeline = pipeline + append_pipeline

        return pipeline
    
    def collect_data(self):
        return super().collect_data()
    
    def query_data(self):
        fetched_data = self.collect_data()
        fetched_data = fetched_data[0]

        process_fn = self.process_function_map.get(self.collection_name, self.process_data_us)
        return process_fn(fetched_data)
    
    def process_data_twse(self, fetched_data):
        """
        處理現金流量表頁面的所有表格
        金流表本身沒有比例 但是Ifa有算，
        項目所屬的情況也不一(分別所屬營業,投資,籌資三個活動)
        所以這裡選擇不用slicing處理
        """

        index_names = []
        column_names = []

        table_dict = {}
        CASHO_dict = {}
        CASHI_dict = {}
        CASHF_dict = {}

        # 處理cash_flow 比例
        checkpoints = ["營業活動之現金流量－間接法", "投資活動之現金流量", "籌資活動之現金流量", "匯率變動對現金及約當現金之影響"]
        main_cash_flows = [
            "營業活動之淨現金流入（流出）", "投資活動之淨現金流入（流出）", "籌資活動之淨現金流入（流出）", "其他"
        ] # 主要的比例對象
        partial_cash_flows = [CASHO_dict, CASHI_dict, CASHF_dict, dict()] 

        # 作法: dictionary中也有checkpoints，如果出現了就換下一個index去計算

        return_dict = {
            "ticker": self.ticker,
            "company_name": fetched_data['company_name']
        }
        seasonal_data = fetched_data.get('seasonal_data')
        if not seasonal_data:
            return_dict.update(self._get_empty_structure())
            return return_dict

        for data in seasonal_data:
            year, season, cash_flow = data['year'], data['season'], data['data']

            time_index = f"{year}Q{season}"

            main_cash_flow_name = ""
            partial_cash_flow = {}
            next_checkpoint = 0

            temp_dict = {}

            for index_name, cash_flow_value in cash_flow.items():
                if (next_checkpoint < 3
                        and index_name == checkpoints[next_checkpoint]): # 找到了主要的變動點
                    main_cash_flow_name = main_cash_flows[next_checkpoint]
                    partial_cash_flow = partial_cash_flows[next_checkpoint]
                    partial_cash_flow[time_index] = {}
                    next_checkpoint += 1

                if main_cash_flow_name: # 有取得cash flow name再進行接下來的計算percentage
                    if (isinstance(cash_flow_value, dict)):
                        value = cash_flow_value.get('value', None) 
                    else:
                        value = cash_flow_value
                    
                    # 處理cashflow percentage部分(取2位 + 接%)
                    main_value = cash_flow.get(main_cash_flow_name, None)
                    if (isinstance(main_value, dict)):
                        main_value = main_value.get('value', None)
                    else:
                        pass

                    try:
                        ratio = np.round(
                            (value / main_value) * 100, 2
                        )
                        ratio = f"{ratio}%"
                    except:
                        ratio = None
                    
                    value = StatsProcessor.cal_non_percentage(value, postfix="千元")
                    temp_dict[index_name] = {
                        "value" : value,
                        "percentage": ratio
                    }
                    partial_cash_flow[time_index][index_name] = temp_dict[index_name]

            table_dict[time_index] = temp_dict
            index_names += list(cash_flow.keys())

        # 轉成dictionary keys
        index_names = list(dict.fromkeys(index_names))

        cash_flow_table = pd.DataFrame(table_dict)
        cash_flow_table_stats = StatsProcessor.expand_value_percentage(cash_flow_table)

        CASHO_table = pd.DataFrame(CASHO_dict)
        CASHO_table = StatsProcessor.expand_value_percentage(CASHO_table)

        CASHI_table = pd.DataFrame(CASHI_dict)
        CASHI_table = StatsProcessor.expand_value_percentage(CASHI_table)

        CASHF_table = pd.DataFrame(CASHF_dict)
        CASHF_table = StatsProcessor.expand_value_percentage(CASHF_table)

        # 回傳歷來格式
        target_season = seasonal_data[0]['season']

        cash_flow_flatten, cash_flow_flatten_YoY = self.flatten_twse(
            table_dict, target_season, index_names
        )

        return_dict.update({
            "cash_flow": cash_flow_table_stats,
            "CASHO": CASHO_table,
            "CASHI": CASHI_table,
            "CASHF": CASHF_table,
            "cash_flow_all": cash_flow_flatten,
            "cash_flow_YoY": cash_flow_flatten_YoY
        })
        return return_dict
    
    def flatten_twse(self, data_dict: dict, target_season: int, index_names: list):
        for time_index in data_dict.keys():
            data_dict[time_index] = self.flatten_dict(data_dict[time_index], index_names, target_keys=['value', 'percentage'])
        cash_flow_flatten = pd.DataFrame.from_dict(data_dict)

        target_season_column = cash_flow_flatten.columns.str.endswith(f"Q{target_season}")

        return cash_flow_flatten, cash_flow_flatten.loc[:, target_season_column]

    def process_data_us(self, fetched_data):

        table_dict = {
            f"{data['year']}Q{data['season']}": data['data'] 
            for data in fetched_data['seasonal_data']
        }

        cash_flow_df = pd.DataFrame.from_dict(table_dict)

        latest_season = fetched_data['seasonal_data'][0]['season']
        target_season_columns = cash_flow_df.columns.str.endswith(
            f"Q{latest_season}"
        )
        cash_flow_df_YoY = cash_flow_df.loc[:, target_season_columns]

        return_dict = {
            "ticker": self.ticker,
            "company_name": fetched_data['company_name'],
            "cash_flow": cash_flow_df,
            "cash_flow_YoY": cash_flow_df_YoY
        }
        return return_dict

    def _get_empty_structure(self):
        return {
            key: pd.DataFrame(columns= pd.Index([], name = 'date')) for key in self.return_keys
        }


from .base import BaseValueTransformer

class TEJValueTransformer(BaseValueTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
    
    def process_transform(self):
        pass


from .base import BaseTEJFetcher
from datetime import datetime
from enum import Enum
import numpy as np
import pandas as pd
from pymongo import MongoClient
from .tech import TechProcessor
from ..utils import StatsProcessor, YoY_Calculator
import warnings
import yaml


class FinanceReportFetcher(BaseTEJFetcher):

    class FetchMode(Enum):
        YOY = 1
        QOQ = 2
        YOY_NOCAL = 3
        QOQ_NOCAL = 4

    def __init__(
        self, mongo_uri, db_name="company", collection_name="TWN/AINVFQ1"
    ):
        self.client = MongoClient(mongo_uri)
        self.db = self.client[db_name]
        self.collection = self.db[collection_name]

        index_files = [
            "tej_db/tej_db_index.yaml", "tej_db/tej_db_thousand_index.yaml",
            "tej_db/tej_db_percent_index.yaml"
        ]

        self.index_dict, self.thousand_dict, self.percent_dict = [
            StatsProcessor.load_yaml(file) for file in index_files
        ]

        self.check_index = set(self.index_dict.get(collection_name, []))
        self.skip_index = set(self.percent_dict.get(collection_name, []))
        self.thousand_index_list = list(
            self.thousand_dict.get(collection_name, [])
        )
        self.percent_index_list = list(
            self.percent_dict.get(collection_name, [])
        )

    def get(
        self,
        ticker,
        fetch_mode: FetchMode = FetchMode.QOQ_NOCAL,
        start_date: str = None,
        end_date: str = None,
        report_type: str = "Q",
        indexes: list = []
    ):
        """
        基礎的query function
        ticker(str): 股票代碼
        start_date(str):  開頭日期範圍
        end_date(str):  = 結束日期範圍
        report_type(str): 報告型態 {"A", "Q", "TTM"}  
        fetch_mode(class FetchMode): 
           YoY : 起始日期到結束日期範圍內，特定該季的資料
           QoQ : 起始日期到結束日期內，每季的資料(與上一季成長率)
        indexes(List): 指定的index
        """
        # 確認indexes中是否有錯誤的index，有的話回傳warning
        if indexes and self.check_index:
            invalid_indexes = set(indexes) - self.check_index
            if invalid_indexes:
                warnings.warn(
                    f"{list(invalid_indexes)} 不存在，請確認欄位名稱", UserWarning
                )

        start_date = datetime.strptime(
            start_date, "%Y-%m-%d"
        ) if start_date else datetime(2005, 1, 1)

        if fetch_mode in {self.FetchMode.QOQ, self.FetchMode.QOQ_NOCAL}:
            end_date = datetime.strptime(end_date, "%Y-%m-%d"
                                         ) if end_date else datetime.today()
            assert start_date <= end_date
            start_year, end_year = start_date.year, end_date.year
            return self.get_QoQ_data(
                ticker=ticker,
                start_date=start_date,
                end_date=end_date,
                report_type=report_type,
                indexes=indexes,
                use_cal=(fetch_mode == self.FetchMode.QOQ)
            )

        elif fetch_mode in {self.FetchMode.YOY, self.FetchMode.YOY_NOCAL}:
            end_date = self.get_latest_data_time(ticker) or datetime.today()
            start_year, end_year = start_date.year, end_date.year
            end_season = (end_date.month - 1) // 4 + 1
            return self.get_YoY_data(
                ticker=ticker,
                start_year=start_year,
                end_year=end_year,
                season=end_season,
                report_type=report_type,
                indexes=indexes,
                use_cal=(fetch_mode == self.FetchMode.YOY)
            )

    def get_QoQ_data(
        self,
        ticker,
        start_date,
        end_date,
        report_type="Q",
        indexes=[],
        use_cal=False
    ):
        """
            取得時間範圍內每季資料
        """
        start_year, start_season = start_date.year, (
            start_date.month - 1
        ) // 4 + 1
        end_year, end_season = end_date.year, (end_date.month - 1) // 4 + 1
        lower_bound_year, lower_bound_season = (
            start_year - 1, 4
        ) if start_season == 1 else (start_year, start_season - 1)

        pipeline = self.build_pipeline(
            ticker, start_year, start_season, end_year, end_season,
            lower_bound_year, lower_bound_season, report_type, indexes
        )
        fetched_data = self.collection.aggregate(pipeline).to_list()
        fetched_data = fetched_data[0]
        fetched_data = fetched_data.get('data', []) if isinstance(fetched_data, dict) else []

        data_dict = self.transform_value(
            StatsProcessor.list_of_dict_to_dict(
                data_list=fetched_data,
                keys=["year", "season"],
                delimeter="Q",
                data_key=report_type
            )
        )

        return self.calculate_and_format(data_dict, use_cal, self.cal_QoQ)

    def get_YoY_data(
        self,
        ticker,
        start_year,
        end_year,
        season,
        report_type="Q",
        indexes=[],
        use_cal=False
    ):
        """
        取得某季歷年資料
        """
        select_year = sorted(
            {year
             for year in range(start_year, end_year + 1)} | {
                 y
                 for year in range(start_year, end_year + 1)
                 for y in {year, year - 1, year - 3, year - 5, year - 10}
             }
        ) if use_cal else list(range(start_year, end_year + 1))

        pipeline = self.build_pipeline(
            ticker,
            select_year,
            season,
            None,
            None,
            None,
            None,
            report_type,
            indexes,
            year_based=True
        )
        fetched_data = self.collection.aggregate(pipeline).to_list()
        fetched_data = fetched_data[0]
        fetched_data = fetched_data.get('data', []) if isinstance(fetched_data, dict) else []

        data_dict = self.transform_value(
            StatsProcessor.list_of_dict_to_dict(
                data_list=fetched_data,
                keys=["year", "season"],
                delimeter="Q",
                data_key=report_type
            )
        )

        return self.calculate_and_format(
            data_dict, use_cal,
            lambda x: self.cal_YoY(x, start_year, end_year, season)
        )

    def transform_value(self, data_dict):
        """
        處理千元, %等單位
        """

        data_df = pd.DataFrame.from_dict(data_dict)
        for category, postfix in [(self.thousand_index_list, "千元"),
                                  (self.percent_index_list, "%")]:
            process_list = list(set(data_df.index) & set(category))
            if postfix == "%":
                data_df = data_df.T
                data_df[process_list] = data_df[process_list].map(lambda x: f"{x}%") # if (not np.isnan(x)) else None)
                data_df = data_df.T
            else:
                data_df.loc[process_list] = data_df.loc[process_list].map(
                    lambda x: StatsProcessor.
                    cal_non_percentage(x, postfix=postfix)
                )
        return data_df.to_dict()

    def build_pipeline(
        self,
        ticker,
        start_year,
        start_season,
        end_year,
        end_season,
        lower_bound_year,
        lower_bound_season,
        report_type,
        indexes,
        year_based=False
    ):

        if year_based:
            filter_cond = {
                "$and": [
                    { "$in": ["$$item.year", start_year] },
                    { "$eq": ["$$item.season", start_season] }
                ]
            }
        else:
            filter_cond = {
                "$or": [
                    {
                        "$and": [
                            { "$gt": ["$$item.year", start_year] },
                            { "$lt": ["$$item.year", end_year] }
                        ]
                    },
                    {
                        "$and": [
                            { "$eq": ["$$item.year", start_year] },
                            { "$gte": ["$$item.season", start_season] }
                        ]
                    },
                    {
                        "$and": [
                            { "$eq": ["$$item.year", end_year] },
                            { "$lte": ["$$item.season", end_season] }
                        ]
                    },
                    {
                        "$and": [
                            { "$eq": ["$$item.year", lower_bound_year] },
                            { "$eq": ["$$item.season", lower_bound_season] }
                        ]
                    }
                ]
            }

        # 每個 filtered item 要輸出哪些欄位
        item_fields = {
            "year": "$$item.year",
            "season": "$$item.season"
        }

        if indexes:
            for idx in indexes:
                item_fields[idx] = f"$$item.{report_type}.{idx}"
        else:
            item_fields[report_type] = f"$$item.{report_type}"
        
        return [
            {
                "$match": {
                    "ticker": ticker
                }
            },
            {
                "$project": {
                    "_id": 0,
                    "ticker": 1,
                    "data": {
                        "$map": {
                            "input": {
                                "$filter": {
                                    "input": "$data",
                                    "as": "item",
                                    "cond": filter_cond
                                }
                            },
                            "as": "item",
                            "in": item_fields
                        }
                    }
                }
            }
        ]

    def calculate_and_format(self, data_dict, use_cal, calc_function):
        data_df = pd.DataFrame.from_dict(
            calc_function(data_dict) if use_cal else data_dict
        ).iloc[:, ::-1]
        return data_df if not use_cal else self.get_dict_of_df(
            data_df.T.to_dict()
        )


class TEJStockPriceFetcher(BaseTEJFetcher):

    def __init__(
        self, mongo_uri, db_name: str = "company", collection_name: str = None
    ):
        self.mongo_uri = mongo_uri
        self.db_name = db_name
        self.collection_name = collection_name

        self.client = MongoClient(self.mongo_uri)
        self.db = self.client[self.db_name]
        self.collection = self.db[self.collection_name]

        self.check_period = [
            '1d', '7d', '1m', '3m', '1y', '3y', '5y', '10y', 'all'
        ]

    def get(
        self,
        ticker: str = "2330",
        start_date: str = "2024-10-01",
        period: str = None
    ):
        """
        取得開高低收資料
        start_date: str: 起始的日期
        period: 指定日期範圍(E.g. 1天, 7天...etc)
        如果宣告period, 以period為優先
        """

        assert (
            period is None or period in self.check_period
        ), f"period should be None or {','.join(self.check_period)}"

        if (period is not None):
            latest_date = self.get_latest_data_time(ticker)
            if (latest_date):
                start_date = self.set_time_shift(date=latest_date, period=period)
            else:
                start_date = datetime.strptime(start_date, "%Y-%m-%d")
        else:
            start_date = datetime.strptime(start_date, "%Y-%m-%d")

        pipeline = [
            {
                "$match": {
                    "ticker": ticker
                }
            },
            {
                "$project": {
                    "_id": 0,
                    "ticker": 1,
                    "data": {
                        "$filter": {
                            "input": "$data",
                            "as": "item",
                            "cond": {
                                "$gte": ["$$item.mdate", start_date]
                            }
                        }
                    }
                }
            }
        ]
        datas = self.collection.aggregate(pipeline).to_list()
        datas = datas[0]
        datas = datas.get('data', [])
        elements = [element for element in datas]
        try:
            data_df = pd.DataFrame(elements).set_index('mdate')
        except:
            column_names = [
                "coid", "mdate", "mkt", "open_d", "high_d", "low_d", "close_d",
                "adjfac", "vol", "amt", "trn", "bid", "offer", "avgprc", "roi",
                "hmlpct", "turnover", "shares", "mktcap", "mktcap_pct",
                "amt_pct", "per", "pbr", "div_yid", "cdiv_yid"
            ]
            data_df = pd.DataFrame(columns = column_names)

        return data_df



from .base import BaseDailyTechDBExtractor
from pymongo import ASCENDING, DESCENDING
from neurostats_API.utils import NotSupportedError

class AsyncDailyValueDBExtractor(BaseDailyTechDBExtractor):

    def __init__(self, ticker, client, fetch_type='D'):
        """
        fetch_type: 
        'D': daily
        'Y': yearly
        """
        self.fetch_type = fetch_type
        super().__init__(ticker, client)
        
    def _get_collection_name(self):
        self.collection_name_map = {
            "tw": {
                "D":"twse_daily_share_price",
                "Y": "twse_yearly_value"
            }
        }
        collection_map = self.collection_name_map.get(self.zone)
        if (collection_map is None):
            raise NotSupportedError(f"AsyncDailyValueDBExtractor only supports {self.collection_name_map.keys}, but get {self.zone}")
        collection_name = collection_map.get(self.fetch_type)
        if (collection_name is None):
            raise ValueError(f"AsyncDailyValueDBExtractor: Invalid fetch type : {self.fetch_type}")
        return collection_name
    
    def _prepare_query(self, start_date = None, end_date = None, get_latest = False):
        query_fn_map = {
            'D': super()._prepare_query,
            'Y': self._prepare_annual_query
        }

        query_fn = query_fn_map.get(self.fetch_type)
        query, projection, sorting = query_fn(start_date=start_date, end_date=end_date, get_latest=get_latest)

        projection.update(
            {   
                "date": 1,
                "year": 1,
                "close": 1,
                "P_E": 1,
                "P_B": 1,
                "P_FCF": 1,
                "P_S": 1,
                "EV_S": 1,
                "EV_OPI": 1,
                "EV_EBIT": 1,
                "EV_EBITDA": 1
            }
        )

        return query, projection, sorting


    def _prepare_annual_query(self, start_date, end_date, get_latest = False):
        start_date = self._transform_date(start_date)
        end_date = self._transform_date(end_date)

        start_year = start_date.year - 1911
        end_year = end_date.year - 1911
        query = {
            "ticker": self.ticker,
            "year": {
                "$gte": start_year,
                "$lte": end_year
            }
        }

        projection = {
            "_id": 0,
        }

        if (get_latest):
            sort = [("year", DESCENDING)]
        else:
            sort = [("year", ASCENDING)]

        return query, projection, sort


class AsyncTEJDailyValueDBExtractor(BaseDailyTechDBExtractor):

    def __init__(self, ticker, mongo_uri):
        super().__init__(ticker, mongo_uri)

    def _get_collection_name(self):
        self.collection_name_map = {"tw": "TEJ_share_price"}

        return self.collection_name_map.get(self.zone, None)



import abc
from datetime import datetime, timedelta, date
import json
import pandas as pd
from pymongo import MongoClient
import pytz
from typing import Union
from neurostats_API.utils import StatsDateTime, StatsProcessor, YoY_Calculator, NotSupportedError

class AsyncBaseFetcher(abc.ABC):
    """
    Fetcher目的: 真正給其他使用者的接口
    """
    def __init__(self):
        self.timezone = pytz.timezone("Asia/Taipei")
        self.target_metric_dict = {
            'value': ['value'],
            'value_and_percentage': ['value', 'percentage'],
            'percentage': ['percentage'],
            'grand_total': ['grand_total'],
            'grand_total_values': ['grand_total', 'grand_total_percentage'],
            'grand_total_percentage': ['grand_total_percentage'],
            'growth': [f'YoY_{i}' for i in [1, 3, 5, 10]],
            'grand_total_growth': [f"YoY_{i}" for i in [1, 3, 5, 10]]
        }
        self.transformer_map = None
    @ abc.abstractmethod
    async def query_data(self):
        """
        user使用的接口
        """
        return NotImplementedError()
    
    @staticmethod
    def str_to_datetime(self, date_str: str) -> StatsDateTime:
        date = self.timezone.localize(datetime.strptime(date_str, "%Y-%m-%d"))
        year, month, day = date.year, date.month, date.day
        season = (month - 1) // 3 + 1
        return StatsDateTime(date, year, month, day, season)
    
    def get_transformer(self, zone):
        transformer = self.transformer_map.get(zone)
        if (transformer is None):
            raise NotSupportedError(
                f"{self.__class__.__name__} get transformer failed: only supports {list(self.transformer_map.keys())}, got {zone}"
            )
        
        return transformer


from .balance_sheet import AsyncBalanceSheetExtractor
from .profit_lose import AsyncProfitLoseExtractor
from .cashflow import AsyncCashFlowExtractor
from .tej import AsyncTEJFinanceStatementDBExtractor, AsyncTEJSelfSettlementDBExtractor


from .base import AsyncBaseFetcher
from ..db_extractors import AsyncBalanceSheetExtractor
from neurostats_API.transformers import (
    TWSEBalanceSheetTransformer, USBalanceSheetTransformer
)
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.utils import NotSupportedError


class AsyncBalanceSheetFetcher(AsyncBaseFetcher):

    def __init__(self, ticker, client):
        super().__init__()
        self.ticker = ticker
        self.transformer_map = {
            "tw": TWSEBalanceSheetTransformer,
            "us": USBalanceSheetTransformer
        }
        self.extractor = get_extractor("DB_balance_sheet", ticker, client)
        company_name = self.extractor.get_company_name()
        zone = self.extractor.get_zone()

        transformer = self.get_transformer(zone)
        self.transformer = transformer(ticker, company_name, zone)

    async def query_data(self, start_date=None, end_date=None):
        fetched_data = await self.extractor.query_data(start_date, end_date)

        transformed_data = self.transformer.process_transform(fetched_data)

        return transformed_data



import abc
from  ..base import BaseTransformer
from neurostats_API.utils import StatsProcessor
import pandas as pd

class BaseMonthRevenueTransformer(BaseTransformer):
    def __init__(self, ticker, company_name, zone):
        self.ticker = ticker
        self.company_name = company_name,
        self.zone = zone

        self.return_keys = []
    
    @abc.abstractmethod
    def process_transform(self):
        pass

    @staticmethod
    def _process_unit(data_df, postfix):

        lambda_map = {
            "revenue":  lambda x: StatsProcessor.cal_non_percentage(x, postfix="千元"),
            "increment": lambda x : StatsProcessor.cal_non_percentage(x, postfix="千元"),
            "ratio": lambda x : StatsProcessor.cal_non_percentage(x, to_str=True, postfix="%"),
            'percentage': lambda x : StatsProcessor.cal_non_percentage(x, to_str=True, postfix="%"),
            "YoY_1": StatsProcessor.cal_percentage,
            "YoY_3": StatsProcessor.cal_percentage,
            "YoY_5": StatsProcessor.cal_percentage,
            "YoY_10": StatsProcessor.cal_percentage
        }

        process_fn = lambda_map.get(postfix)
        postfix_cols = data_df.columns.str.endswith(postfix)
        postfix_cols = data_df.loc[:, postfix_cols].columns

        data_df[postfix_cols] = (
            data_df[postfix_cols].map(
               process_fn
            )
        )

        return data_df
    
    def _apply_process_unit_pipeline(
        self,
        data_df,
        postfix_list = ['revenue', 'increment']
    ):
        return super()._apply_process_unit_pipeline(
            data_df,
            postfix_list
        )

    def _process_data(self, fetched_data):
        for data in fetched_data:
            data.pop("ticker")
            data.pop("company_name")
            data.pop("memo")
        
        return pd.DataFrame(fetched_data)


from .base import BaseChipTransformer
from datetime import date, datetime
import inspect
import numpy as np
import pandas as pd
from neurostats_API.utils import StatsProcessor
from collections import defaultdict
from typing import List, Optional, Union

class US_F13_Transformer(BaseChipTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
        self.return_keys = ['chip_data']

    def process_transform(self, fetched_datas):
        """
        fetched_datas:[
            {

            }, ....
        ]
        """
        if not fetched_datas:
            return self._get_empty_structure()    # 若查無資料，回傳空表結構

        def flatten_nested_data(data_dict, key):
            pop_dict = data_dict.pop(key, {})
            for sub_key, value in pop_dict.items():
                data_dict[f"{key}_{sub_key}"] = value
            
            return data_dict

        result = [flatten_nested_data(data, 'votingAuthority') for data in fetched_datas]

        return_dict = {
            "ticker": self.ticker,
            "company_name": self.company_name,
            "chip_data": pd.DataFrame(result)
        }

        return return_dict

class US_CelebrityTransformer():

    def __init__(self):
        pass

    def process_transform(
        self,
        fetched_datas: List[dict],
        strftime: Optional[str] = None
    ) -> dict[str, dict[Union[str, datetime], float]]:
        
        result = defaultdict(lambda: defaultdict(dict))

        for record in fetched_datas:
            ticker = record.get("issuerTicker")
            report_date = record.get("reportDate")
            value = record.get("value")
            title = record.get("titleOfClass")

            if ticker and report_date and title:
                key = report_date.strftime(strftime) if strftime else report_date
                result.setdefault(ticker, {}).setdefault(title, {})[key] = value

        return dict(result)


from datetime import datetime

from neurostats_API.async_mode.fetchers.base import AsyncBaseFetcher

from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import TEJChipTransformer
from neurostats_API.utils import NotSupportedError

class AsyncTEJChipFetcher(AsyncBaseFetcher):
    def __init__(
        self,
        ticker,
        client
    ):
        self.ticker = ticker
        self.transformer_map = {"tw": TEJChipTransformer}
        try:
            self.extractor = get_extractor("TEJ_Chip", ticker, client)

            company_name = self.extractor.get_company_name()
            zone = self.extractor.get_zone()

            transformer = self.get_transformer(zone)
            self.transformer = transformer(ticker, company_name, zone)

        except NotSupportedError as e:
            raise NotSupportedError(
                f"{self.__class__.__name__} only support {list(self.transformer_map.keys())} companies now, {ticker} is not available"
            ) from e

    async def query_data(
        self, start_date=None, end_date=None, get_latest=False
    ):
        fetched_data = await self.extractor.query_data(
            start_date, end_date, get_latest
        )

        transformed_data = self.transformer.process_transform(
            fetched_data = fetched_data
        )

        return transformed_data



from .base import StatsFetcher, StatsDateTime
import json
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor
import yaml


class BalanceSheetFetcher(StatsFetcher):
    """
    對應iFa.ai -> 財務分析 ->  資產負債表
    """

    def __init__(self, ticker, db_client):
        super().__init__(ticker, db_client)
        self.table_settings = StatsProcessor.load_yaml("twse/balance_sheet.yaml")

        self.process_function_map = {
            "twse_stats": self.process_data_twse,
            "us_stats": self.process_data_us
        }

        self.return_keys = [
            'balance_sheet', 'total_asset', 'current_asset', 'non_current_asset', 
            'current_debt', 'non_current_debt', 'equity', 'balance_sheet_all', 'balance_sheet_YoY'
        ]

    def prepare_query(self):
        pipeline = super().prepare_query()

        name_map = {
            "twse_stats": "balance_sheet",
            "us_stats": "balance_sheet"
        }


        chart_name = name_map.get(self.collection_name, "balance_sheet")

        append_pipeline = [
        {
            "$project": {
                "_id": 0,
                "ticker": 1,
                "company_name": 1,
                "seasonal_data": {
                    "$map": {
                        "input": {"$ifNull": ["$seasonal_data", []]},
                        "as": "season",
                        "in": {
                            "year": "$$season.year",
                            "season": "$$season.season",
                            "data": {"$ifNull": [f"$$season.{chart_name}", []]}
                        }
                    }
                }
            }
        }
    ]
        pipeline = pipeline + append_pipeline

        return pipeline

    def collect_data(self):
        return super().collect_data()

    def query_data(self):
        fetched_data = self.collect_data()
        fetched_data = fetched_data[0]

        process_fn = self.process_function_map[self.collection_name]
        processed_data = process_fn(fetched_data)
        return processed_data

    def process_data_twse(self, fetched_data):
        latest_time = StatsDateTime.get_latest_time(
            self.ticker, self.collection
        ).get('last_update_time', {})
        # 取最新時間資料時間，沒取到就預設去年年底
        target_year = latest_time.get('seasonal_data', {}).get(
            'latest_target_year',
            StatsDateTime.get_today().year - 1
        )
        target_season = latest_time.get('seasonal_data',
                                        {}).get('latest_season', 4)

        return_dict = {
            "ticker": self.ticker,
            "company_name": fetched_data['company_name']
        }
        table_dict = {}

        seasonal_data = fetched_data.get('seasonal_data')
        if not seasonal_data:
            return_dict.update(self._get_empty_structure())
            return return_dict

        # 將value與percentage跟著年分季度一筆筆取出
        for data in seasonal_data:
            year, season, balance_sheet = data['year'], data['season'], data[
                'data']
            time_index = f"{year}Q{season}"

            new_balance_sheet = dict()
            # 蒐集整體的keys
            index_names = list(balance_sheet.keys())

            table_dict[time_index] = balance_sheet
            # flatten dict:
            # {<key>: {"value": <value>, "percentage": <value>}}
            # -> {<key>_value: <value>, <key>_percentage:<value>}
        
        old_balance_sheet = pd.DataFrame(table_dict)
        target_season_col = old_balance_sheet.columns.str.endswith(f"Q{target_season}")
        old_balance_sheet = old_balance_sheet.loc[:, target_season_col]
        old_balance_sheet = StatsProcessor.expand_value_percentage(old_balance_sheet)

        # 處理QoQ版BalanceSheet
        for time_index, data_dict in table_dict.items():
            new_balance_sheet = self.flatten_dict(
                data_dict, 
                indexes = index_names,
                target_keys=["value", "percentage"]
            )

            table_dict[time_index] = new_balance_sheet

        total_table = pd.DataFrame.from_dict(table_dict).T
        value_index = total_table.columns.str.endswith("_value")
        value_cols = total_table.loc[:, value_index].columns
        total_table[value_cols] = (
            total_table[value_cols].map(
                lambda x: StatsProcessor.cal_non_percentage(x, postfix="千元"),
            )
        )

        percentage_index = total_table.columns.str.endswith(
            "_percentage"
        )
        percentage_cols = total_table.loc[:, percentage_index].columns
        total_table[percentage_cols] = (
            total_table[percentage_cols].map(
                lambda x: StatsProcessor.
                cal_non_percentage(x, to_str=True, postfix="%"),
            )
        )

        total_table = total_table.T
        target_season_columns = total_table.columns.str.endswith(
            f"Q{target_season}"
        )
        total_table_YoY = total_table.loc[:, target_season_columns]

        for name, setting in self.table_settings.items():
            target_indexes = setting.get('target_index', [None])
            for target_index in target_indexes:
                try:
                    return_dict[name] = StatsProcessor.slice_old_table(
                        total_table=old_balance_sheet,
                        target_index=target_index
                    )
                    break
                except Exception as e:
                    continue
        
        return_dict.update(
            {
                "balance_sheet": old_balance_sheet,
                "balance_sheet_all": total_table.copy(),
                "balance_sheet_YoY": total_table_YoY
            }
        )
        return return_dict

    def process_data_us(self, fetched_data):
        return_dict = {
            "ticker": self.ticker,
            "company_name": fetched_data['company_name']
        }
        table_dict = {}

        table_dict = dict()

        for data in fetched_data['seasonal_data']:
            year, season, balance_sheet = data['year'], data['season'], data[
                'data']
            table_dict[f"{year}Q{season}"] = balance_sheet

        table_dict = pd.DataFrame.from_dict(table_dict)

        return_dict["balance_sheet"] = table_dict

        latest_season = fetched_data['seasonal_data'][0]['season']
        target_season_columns = table_dict.columns.str.endswith(
            f"Q{latest_season}"
        )
        table_dict_YoY = table_dict.loc[:, target_season_columns]
        return_dict["balance_sheet_YoY"] = table_dict_YoY
        return return_dict

    def _get_empty_structure(self):
        return {
            key: pd.DataFrame(columns= pd.Index([], name = 'date')) for key in self.return_keys
        }


from .base import BaseProfitLoseTransformer
from neurostats_API.utils import StatsProcessor, YoY_Calculator
import pandas as pd


class TWSEProfitLoseTransformer(BaseProfitLoseTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

        # 載入欄位切片設定檔（定義哪些欄位要顯示在報表上）
        self.table_settings = StatsProcessor.load_yaml("twse/profit_lose.yaml")

        # 設定最後回傳的所有欄位 key，包含主頁面、總覽、成長率、比率等
        self.return_keys = [
            'profit_lose', 'grand_total_profit_lose', 'revenue', 'grand_total_revenue', 
            'gross_profit', 'grand_total_gross_profit', 'gross_profit_percentage', 
            'grand_total_gross_profit_percentage', 'operating_income', 'grand_total_operating_income', 'operating_income_percentage', 
            'grand_total_operating_income_percentage', 'net_income_before_tax', 'grand_total_net_income_before_tax', 'net_income_before_tax_percentage', 
            'grand_total_net_income_before_tax_percentage', 'net_income', 'grand_total_net_income', 'net_income_percentage', 
            'grand_total_income_percentage', 'EPS', 'EPS_growth', 'grand_total_EPS', 
            'grand_total_EPS_growth', 'profit_lose_all', 'profit_lose_YoY'
        ]

        self.stats_df = None  # 主頁面統計用 DataFrame
        self.new_df = None    # 完整表格用 DataFrame（包含 TEJ 格式）
    
    def process_transform(self, fetched_data):
        if (not fetched_data):
            return self._get_empty_structure()  # 若查無資料，回傳空表結構

        processed_data = self._process_fn(fetched_data)

        return processed_data


    def _process_fn(self, fetched_data):
        return_dict = {
            "ticker": self.ticker,
            "company_name": self.company_name,
        }

        target_season = fetched_data[-1]['season']  # 用來選擇同一季的欄位當作 YoY 主體

        # 轉換原始 JSON -> dict 格式（key: Q1 ~ Q4）
        stats_dict = self._process_twse_to_stats_format(fetched_data).to_dict()

        # 計算 QoQ 與 YoY 成長率欄位
        stats_dict = YoY_Calculator.cal_QoQ(stats_dict)

        # 建立主頁面 DataFrame 並擴充 _value / _percentage 欄位
        self.stats_df = pd.DataFrame.from_dict(stats_dict)
        self.stats_df = self._slice_target_season(self.stats_df, target_season)
        self.stats_df = StatsProcessor.expand_value_percentage(self.stats_df)

        # 處理各種單位轉換欄位
        self.stats_df = self._apply_process_unit_pipeline(
            self.stats_df,
            postfix_list=[
                '_value',
                '_percentage',
                '_growth',
                "_YoY_1",
                "_YoY_3",
                "_YoY_5",
                "_YoY_10",
            ]
        )

        # 篩選主頁面欄位（排除 grand_total 欄）
        grand_total_value_col = self.stats_df.columns.str.endswith("grand_total_value")
        grand_total_percentage_col = self.stats_df.columns.str.endswith("grand_total_percentage")
        self.stats_df = self.stats_df.loc[
            :,
            (~grand_total_value_col) & (~grand_total_percentage_col) 
        ]

        
        stats_main_page_df = StatsProcessor.slice_table(
            self.stats_df,
            mode = 'value_and_percentage',
        )
        stats_grand_total_df = StatsProcessor.slice_table(
            self.stats_df,
            mode = 'grand_total_values'
        )

        # 建立完整表格 new_df，轉為 TEJ 標準格式並做單位處理
        self.new_df = self._process_twse_to_tej_format(stats_dict)
        self.new_df = self._apply_process_unit_pipeline(
            self.new_df,
            postfix_list=[
                '_value',
                '_percentage',
                '_growth',
                "_YoY_1",
                "_YoY_3",
                "_YoY_5",
                "_YoY_10",
            ]
        )
        # 建立 YoY 子表格（以 transposed 的方式）
        new_df_YoY = self._slice_target_season(self.new_df.T, target_season)

        # 回填主欄位資訊至結果 dict
        return_dict.update({
            "profit_lose": stats_main_page_df,
            "grand_total_profit_lose": stats_grand_total_df,
            "profit_lose_all": self.new_df,
            "profit_lose_YoY": new_df_YoY
        })

        # 根據設定檔切出其他欄位（如營收、毛利率等）
        self._process_target_columns(
            return_dict,
            new_df_YoY.T,
        )

        return return_dict

    def _process_target_columns(self, return_dict ,data_df):
        for name, setting in self.table_settings.items():
            target_indexes = setting.get('target_index', [None])
            for target_index in target_indexes:
                try:
                    # 使用設定檔定義的 mode + index，切出對應欄位表格
                    return_dict[name] = StatsProcessor.slice_table(
                        total_table=data_df,
                        mode=setting['mode'],
                        target_index=target_index
                    )
                    break  # 第一個成功就跳出（允許 fallback 多個 index）
                except Exception as e:
                    continue


from .base import AsyncBaseFetcher
from datetime import datetime, timedelta
from neurostats_API.async_mode.factory import get_extractor
from ..db_extractors import (
    AsyncTWSEChipDBExtractor,
    AsyncYFDailyTechDBExtractor,
)

from neurostats_API.transformers import TWSEChipTransformer
from neurostats_API.utils import NotSupportedError
import pandas as pd


class AsyncTWSEMarginFetcher(AsyncBaseFetcher):
    """
    stats 網頁用
    """

    def __init__(self, ticker, client):
        self.ticker = ticker
        self.transformer_map = {"tw": TWSEChipTransformer}
        try:
            self.margin_extractor = get_extractor(
                "TWSE_chip", ticker, client, fetch_type='M'
            )
            self.security_extractor = get_extractor(
                "TWSE_chip", ticker, client, fetch_type='S'
            )
            self.daily_extractor = get_extractor("YF_tech", ticker, client)

            company_name = self.margin_extractor.get_company_name()
            zone = self.margin_extractor.get_zone()



            transformer = self.get_transformer(zone)
            self.transformer = transformer(ticker, company_name, zone)
        except NotSupportedError as e:
            raise NotSupportedError(
                f"{self.__class__.__name__} only support {list(self.transformer_map.keys())} companies now, {ticker} is not available"
            ) from e

    async def query_data(self, start_date=None, end_date=None):
        margin = await self.margin_extractor.query_data(start_date, end_date)
        security_lending = await self.security_extractor.query_data(
            start_date, end_date
        )

        tech_data = await self.daily_extractor.query_data(start_date, end_date)

        fetched_data = {
            "margin_trading": margin,
            "security_lending": security_lending
        }
        transformed_data = self.transformer.process_transform(
            tech_data=tech_data, **fetched_data
        )
        transformed_data['annual_trading'] = self._concat_data(
            transformed_data['annual_trading']
        )

        return transformed_data

    async def query_data_annual(self):
        """
        取一年內的資料
        """
        end_date = datetime.today()
        start_date = end_date - timedelta(days=365)

        fetched_data = await self.query_data(start_date, end_date)

        return fetched_data

    def _concat_data(self, annaual_data):
        """
        借券, 債券串接
        annaual_data: 
        {
            "margin_trading": pd.DataFrame,
            'security_lending': pd.DataFrame
        }
        """

        margin_annual = annaual_data.get("margin_trading")
        security_annaul = annaual_data.get("security_lending")

        unique_cols = [
            name for name in security_annaul.columns
            if name not in margin_annual.columns
        ]
        concat_df = pd.concat(
            [margin_annual, security_annaul[unique_cols]], axis=1
        )

        return {
            'margin_trading': margin_annual,
            'security_lending': security_annaul,
            'stats_page_history': concat_df
        }



from neurostats_API.async_mode.db_extractors import AsyncProfitLoseExtractor
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import (
    TWSEProfitLoseTransformer,
    USProfitLoseTransformer
)
from neurostats_API.utils import (
    NoCompanyError,
    NotSupportedError
)

from .base import AsyncBaseFetcher

class AsyncProfitLoseFetcher(AsyncBaseFetcher):
    def __init__(
        self, 
        ticker,
        client
    ):
        self.ticker = ticker
        
        self.transformer_map = {
            "tw": TWSEProfitLoseTransformer,
            "us": USProfitLoseTransformer
        }
        try:
            self.extractor = get_extractor("DB_profit_lose", ticker, client)
            company_name = self.extractor.get_company_name()
            zone = self.extractor.get_zone()

            transformer = self.get_transformer(zone)
            self.transformer = transformer(ticker, company_name, zone)
        
        except NotSupportedError as e:
            raise NotSupportedError(
                 f"{self.__class__.__name__} only support {list(self.transformer_map.keys())} companies now, {ticker} is not available"
            ) from e

    async def query_data(self, start_date = None, end_date = None):
        fetched_data = await self.extractor.query_data(start_date, end_date)

        transformed_data = self.transformer.process_transform(
            fetched_data
        )

        return transformed_data


from .base import AsyncBaseSeasonalDBExtractor

class AsyncCashFlowExtractor(AsyncBaseSeasonalDBExtractor):
    def __init__(self, ticker, client):
        super().__init__(ticker, client)

        self.column_name_map = {
            'tw': "cash_flow",
            'us': "cash_flow"
        }


from .fetchers import (
    AsyncAgentOverviewFetcher,
    AsyncBalanceSheetFetcher,
    AsyncCashFlowFetcher,
    AsyncFinanceOverviewFetcher,
    AsyncMonthlyRevenueFetcher,
    AsyncProfitLoseFetcher,
    AsyncTEJChipFetcher,
    AsyncTechFetcher,
    AsyncTEJSeasonalFetcher,
    AsyncTWSEInstitutionFetcher,
    AsyncTWSEMarginFetcher,
    AsyncTWSEStatsValueFetcher,
    AsyncUSCelebrityFetcher,
    AsyncBatchTechFetcher
)


from .base import AsyncBaseSeasonalDBExtractor
from datetime import datetime
import json
import pandas as pd
from pymongo import ASCENDING, DESCENDING
from neurostats_API.async_mode.db import TWSEDBClient, USDBClient
from neurostats_API.utils import StatsDateTime, StatsProcessor
import yaml

class AsyncBalanceSheetExtractor(AsyncBaseSeasonalDBExtractor):

    def __init__(self, ticker, client):

        super().__init__(ticker, client)
        
        self.column_name_map = {
            'tw': "balance_sheet",
            'us': "balance_sheet"
        }